from abc import abstractmethod, ABC

from cg.services.deliver_files.file_fetcher.models import DeliveryFiles


class FilterDeliveryFilesService(ABC):

    @abstractmethod
    def filter_delivery_files(self, delivery_files: DeliveryFiles, sample_id: str) -> DeliveryFiles:
        pass



from cg.apps.lims import LimsAPI
from cg.models.orders.constants import OrderType
from cg.services.orders.order_lims_service.order_lims_service import OrderLimsService
from cg.services.orders.store_order_services.store_case_order import (
    StoreCaseOrderService,
)
from cg.services.orders.store_order_services.store_fastq_order_service import (
    StoreFastqOrderService,
)
from cg.services.orders.store_order_services.store_metagenome_order import (
    StoreMetagenomeOrderService,
)
from cg.services.orders.store_order_services.store_microbial_fastq_order_service import (
    StoreMicrobialFastqOrderService,
)
from cg.services.orders.store_order_services.store_microbial_order import (
    StoreMicrobialOrderService,
)
from cg.services.orders.store_order_services.store_pacbio_order_service import (
    StorePacBioOrderService,
)
from cg.services.orders.store_order_services.store_pool_order import (
    StorePoolOrderService,
)
from cg.services.orders.submitters.case_order_submitter import CaseOrderSubmitter
from cg.services.orders.submitters.fastq_order_submitter import FastqOrderSubmitter
from cg.services.orders.submitters.metagenome_order_submitter import (
    MetagenomeOrderSubmitter,
)
from cg.services.orders.submitters.microbial_order_submitter import (
    MicrobialOrderSubmitter,
)
from cg.services.orders.submitters.order_submitter import OrderSubmitter
from cg.services.orders.submitters.pacbio_order_submitter import PacbioOrderSubmitter
from cg.services.orders.submitters.pool_order_submitter import PoolOrderSubmitter
from cg.services.orders.validate_order_services.validate_case_order import (
    ValidateCaseOrderService,
)
from cg.services.orders.validate_order_services.validate_fastq_order import (
    ValidateFastqOrderService,
)
from cg.services.orders.validate_order_services.validate_metagenome_order import (
    ValidateMetagenomeOrderService,
)
from cg.services.orders.validate_order_services.validate_microbial_order import (
    ValidateMicrobialOrderService,
)
from cg.services.orders.validate_order_services.validate_pacbio_order import (
    ValidatePacbioOrderService,
)
from cg.services.orders.validate_order_services.validate_pool_order import (
    ValidatePoolOrderService,
)
from cg.store.store import Store


class OrderSubmitterRegistry:
    """
    A registry for OrderSubmitter instances, keyed by OrderType.
    """

    def __init__(self):
        self._registry = {}

    def register(self, order_type: OrderType, order_submitter: OrderSubmitter):
        """Register an OrderSubmitter instance for a given OrderType."""
        self._registry[order_type] = order_submitter

    def get_order_submitter(self, order_type: OrderType) -> OrderSubmitter:
        """Fetch the registered OrderSubmitter for the given OrderType."""
        if order_submitter := self._registry.get(order_type):
            return order_submitter
        raise ValueError(f"No OrderSubmitter registered for order type: {order_type}")


order_service_mapping = {
    OrderType.BALSAMIC: (
        OrderLimsService,
        ValidateCaseOrderService,
        StoreCaseOrderService,
        CaseOrderSubmitter,
    ),
    OrderType.BALSAMIC_QC: (
        OrderLimsService,
        ValidateCaseOrderService,
        StoreCaseOrderService,
        CaseOrderSubmitter,
    ),
    OrderType.BALSAMIC_UMI: (
        OrderLimsService,
        ValidateCaseOrderService,
        StoreCaseOrderService,
        CaseOrderSubmitter,
    ),
    OrderType.FASTQ: (
        OrderLimsService,
        ValidateFastqOrderService,
        StoreFastqOrderService,
        FastqOrderSubmitter,
    ),
    OrderType.FLUFFY: (
        OrderLimsService,
        ValidatePoolOrderService,
        StorePoolOrderService,
        PoolOrderSubmitter,
    ),
    OrderType.METAGENOME: (
        OrderLimsService,
        ValidateMetagenomeOrderService,
        StoreMetagenomeOrderService,
        MetagenomeOrderSubmitter,
    ),
    OrderType.MICROBIAL_FASTQ: (
        OrderLimsService,
        ValidateMicrobialOrderService,
        StoreMicrobialFastqOrderService,
        MicrobialOrderSubmitter,
    ),
    OrderType.MICROSALT: (
        OrderLimsService,
        ValidateMicrobialOrderService,
        StoreMicrobialOrderService,
        MicrobialOrderSubmitter,
    ),
    OrderType.MIP_DNA: (
        OrderLimsService,
        ValidateCaseOrderService,
        StoreCaseOrderService,
        CaseOrderSubmitter,
    ),
    OrderType.MIP_RNA: (
        OrderLimsService,
        ValidateCaseOrderService,
        StoreCaseOrderService,
        CaseOrderSubmitter,
    ),
    OrderType.PACBIO_LONG_READ: (
        OrderLimsService,
        ValidatePacbioOrderService,
        StorePacBioOrderService,
        PacbioOrderSubmitter,
    ),
    OrderType.RML: (
        OrderLimsService,
        ValidatePoolOrderService,
        StorePoolOrderService,
        PoolOrderSubmitter,
    ),
    OrderType.RNAFUSION: (
        OrderLimsService,
        ValidateCaseOrderService,
        StoreCaseOrderService,
        CaseOrderSubmitter,
    ),
    OrderType.SARS_COV_2: (
        OrderLimsService,
        ValidateMicrobialOrderService,
        StoreMicrobialOrderService,
        MicrobialOrderSubmitter,
    ),
    OrderType.TAXPROFILER: (
        OrderLimsService,
        ValidateMetagenomeOrderService,
        StoreMetagenomeOrderService,
        MetagenomeOrderSubmitter,
    ),
    OrderType.TOMTE: (
        OrderLimsService,
        ValidateCaseOrderService,
        StoreCaseOrderService,
        CaseOrderSubmitter,
    ),
}


def build_submitter(lims: LimsAPI, status_db: Store, order_type: OrderType) -> OrderSubmitter:
    """Build an OrderSubmitter instance for the given OrderType."""
    lims_service, validation_service, store_service, submitter_class = order_service_mapping[
        order_type
    ]
    return submitter_class(
        order_validation_service=validation_service(status_db),
        order_store_service=store_service(status_db, lims_service(lims)),
    )


def setup_order_submitter_registry(lims: LimsAPI, status_db: Store) -> OrderSubmitterRegistry:
    """Set up the OrderSubmitterRegistry with all OrderSubmitter instances."""
    registry = OrderSubmitterRegistry()
    for order_type in order_service_mapping.keys():
        registry.register(
            order_type=order_type,
            order_submitter=build_submitter(lims=lims, status_db=status_db, order_type=order_type),
        )
    return registry



"""Module to create delivery reports."""

import logging
from datetime import datetime
from pathlib import Path
from typing import Any

from housekeeper.store.models import File, Version
from jinja2 import Environment, PackageLoader, Template, select_autoescape
from sqlalchemy.orm import Query

from cg.apps.coverage import ChanjoAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.constants import DELIVERY_REPORT_FILE_NAME, SWEDAC_LOGO_PATH, Workflow
from cg.constants.constants import MAX_ITEMS_TO_RETRIEVE, FileFormat
from cg.constants.housekeeper_tags import HK_DELIVERY_REPORT_TAG, HermesFileTag
from cg.constants.scout import ScoutUploadKey
from cg.exc import DeliveryReportError
from cg.io.controller import ReadFile, WriteStream
from cg.meta.delivery.delivery import DeliveryAPI
from cg.meta.delivery_report.data_validators import get_empty_report_data, get_missing_report_data
from cg.meta.workflow.analysis import AnalysisAPI
from cg.models.analysis import AnalysisModel
from cg.models.delivery_report.metadata import SampleMetadataModel
from cg.models.delivery_report.report import (
    CaseModel,
    CustomerModel,
    DataAnalysisModel,
    ReportModel,
    ScoutVariantsFiles,
)
from cg.models.delivery_report.sample import (
    ApplicationModel,
    MethodsModel,
    SampleModel,
    TimestampModel,
)
from cg.store.models import (
    Analysis,
    Application,
    ApplicationLimitations,
    Case,
    CaseSample,
    Sample,
)
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class DeliveryReportAPI:
    """Common delivery report API."""

    def __init__(self, analysis_api: AnalysisAPI):
        self.analysis_api: AnalysisAPI = analysis_api
        self.chanjo_api: ChanjoAPI = self.analysis_api.chanjo_api
        self.delivery_api: DeliveryAPI = self.analysis_api.delivery_api
        self.housekeeper_api: HousekeeperAPI = self.analysis_api.housekeeper_api
        self.lims_api: LimsAPI = self.analysis_api.lims_api
        self.scout_api: ScoutAPI = self.analysis_api.scout_api
        self.status_db: Store = self.analysis_api.status_db

    def get_delivery_report_html(self, case_id: str, analysis_date: datetime, force: bool) -> str:
        """Generates the HTML content of a delivery report."""
        report_data: ReportModel = self.get_report_data(
            case_id=case_id, analysis_date=analysis_date
        )
        report_data: ReportModel = self.validate_report_data(
            case_id=case_id, report_data=report_data, force=force
        )
        rendered_report: str = self.render_delivery_report(report_data=report_data.model_dump())
        return rendered_report

    def write_delivery_report_file(
        self, case_id: str, directory: Path, analysis_date: datetime, force: bool
    ) -> Path:
        """Write a file containing the delivery report content."""
        directory.mkdir(parents=True, exist_ok=True)
        delivery_report: str = self.get_delivery_report_html(
            case_id=case_id, analysis_date=analysis_date, force=force
        )
        report_file_path = Path(directory, DELIVERY_REPORT_FILE_NAME)
        with open(report_file_path, "w") as delivery_report_stream:
            delivery_report_stream.write(delivery_report)
        return report_file_path

    def add_delivery_report_to_hk(
        self, case_id: str, delivery_report_file: Path, version: Version
    ) -> File | None:
        """Add a delivery report file to a case bundle and return its file object."""
        LOG.info(f"Adding a new delivery report to Housekeeper for {case_id}")
        file: File = self.housekeeper_api.add_file(
            path=delivery_report_file.as_posix(),
            version_obj=version,
            tags=[
                case_id,
                HK_DELIVERY_REPORT_TAG,
                HermesFileTag.CLINICAL_DELIVERY,
                HermesFileTag.LONG_TERM_STORAGE,
            ],
        )
        self.housekeeper_api.include_file(file, version)
        self.housekeeper_api.add_commit(file)
        return file

    def get_delivery_report_from_hk(self, case_id: str, version: Version) -> str | None:
        """Return path of a delivery report stored in HK."""
        delivery_report: File = self.housekeeper_api.get_latest_file(
            bundle=case_id, tags=[HK_DELIVERY_REPORT_TAG], version=version.id
        )
        if not delivery_report:
            LOG.warning(f"No delivery report found in housekeeper for {case_id}")
            return None
        return delivery_report.full_path

    def get_scout_uploaded_file_from_hk(
        self, case_id: str, scout_key: ScoutUploadKey
    ) -> str | None:
        """Return file path of the uploaded to Scout file given its tag."""
        version: Version = self.housekeeper_api.last_version(bundle=case_id)
        tags: list = self.get_hk_scout_file_tags(scout_key=scout_key)
        uploaded_file: File = self.housekeeper_api.get_latest_file(
            bundle=case_id, tags=tags, version=version.id
        )
        if not tags or not uploaded_file:
            LOG.warning(f"No files were found for the following Scout key: {scout_key}")
            return None
        return uploaded_file.full_path

    @staticmethod
    def render_delivery_report(report_data: dict) -> str:
        """Renders the report on the Jinja template."""
        env = Environment(
            loader=PackageLoader("cg", Path("meta", "delivery_report", "templates").as_posix()),
            autoescape=select_autoescape(["html", "xml"]),
        )
        env.globals["get_content_from_file"] = ReadFile.get_content_from_file
        env.globals["swedac_logo_path"] = SWEDAC_LOGO_PATH
        template: Template = env.get_template(name=DELIVERY_REPORT_FILE_NAME)
        return template.render(**report_data)

    def get_cases_without_delivery_report(self, workflow: Workflow) -> list[Case]:
        """Returns a list of cases that has been stored and need a delivery report."""
        stored_cases: list[Case] = []
        analyses: Query = self.status_db.analyses_to_delivery_report(workflow=workflow)[
            :MAX_ITEMS_TO_RETRIEVE
        ]
        for analysis in analyses:
            case: Case = analysis.case
            last_version: Version = self.housekeeper_api.last_version(bundle=case.internal_id)
            hk_file: File = self.housekeeper_api.get_files(
                bundle=case.internal_id, version=last_version.id if last_version else None
            ).first()

            if hk_file and Path(hk_file.full_path).is_file():
                stored_cases.append(case)
            else:
                LOG.warning(
                    f"Case {case.internal_id} must be stored before creating a delivery report"
                )
        return stored_cases

    def get_cases_without_uploaded_delivery_report(self, workflow: Workflow) -> list[Case]:
        """Returns a list of cases that need a delivery report to be uploaded."""
        analyses: Query = self.status_db.analyses_to_upload_delivery_reports(workflow=workflow)[
            :MAX_ITEMS_TO_RETRIEVE
        ]
        return [analysis.case for analysis in analyses]

    def update_delivery_report_date(self, case: Case, analysis_date: datetime) -> None:
        """Updates the date when a delivery report was created."""
        analysis: Analysis = self.status_db.get_analysis_by_case_entry_id_and_started_at(
            case_entry_id=case.id, started_at_date=analysis_date
        )
        analysis.delivery_report_created_at = datetime.now()
        self.status_db.session.commit()

    def get_report_data(self, case_id: str, analysis_date: datetime) -> ReportModel:
        """Fetches all the data needed to generate a delivery report."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        analysis: Analysis = self.status_db.get_analysis_by_case_entry_id_and_started_at(
            case_entry_id=case.id, started_at_date=analysis_date
        )
        analysis_metadata: AnalysisModel = self.analysis_api.get_latest_metadata(case.internal_id)
        case_model: CaseModel = self.get_case_data(case, analysis, analysis_metadata)
        return ReportModel(
            customer=self.get_customer_data(case=case),
            version=self.get_report_version(analysis=analysis),
            date=datetime.today(),
            case=case_model,
            accredited=self.is_report_accredited(
                samples=case_model.samples, analysis_metadata=analysis_metadata
            ),
        )

    def validate_report_data(
        self, case_id: str, report_data: ReportModel, force: bool
    ) -> ReportModel:
        """Verifies that the required report fields are not empty."""
        required_fields: dict = self.get_required_fields(case=report_data.case)
        empty_report_fields: dict = get_empty_report_data(report_data=report_data)
        missing_report_fields: dict = get_missing_report_data(
            empty_fields=empty_report_fields, required_fields=required_fields
        )
        if missing_report_fields and not force:
            LOG.error(
                f"Could not generate report data for {case_id}. "
                f"Missing data: \n{WriteStream.write_stream_from_content(content=missing_report_fields, file_format=FileFormat.YAML)}"
            )
            raise DeliveryReportError
        if empty_report_fields:
            LOG.warning(
                f"Empty report fields: \n{WriteStream.write_stream_from_content(content=empty_report_fields, file_format=FileFormat.YAML)}"
            )
        return report_data

    @staticmethod
    def get_customer_data(case: Case) -> CustomerModel:
        """Returns customer validated attributes retrieved from status DB."""
        return CustomerModel(
            name=case.customer.name,
            id=case.customer.internal_id,
            invoice_address=case.customer.invoice_address,
            scout_access=case.customer.scout_access,
        )

    @staticmethod
    def get_report_version(analysis: Analysis) -> int:
        """
        Returns the version of the given analysis. The version of the first analysis is 1 and subsequent reruns
        increase it by 1.
        """
        version = None
        if analysis:
            version = len(analysis.case.analyses) - analysis.case.analyses.index(analysis)
        return version

    def get_case_data(
        self,
        case: Case,
        analysis: Analysis,
        analysis_metadata: AnalysisModel,
    ) -> CaseModel:
        """Returns case associated validated attributes."""
        samples: list[SampleModel] = self.get_samples_data(
            case=case, analysis_metadata=analysis_metadata
        )
        unique_applications: list[ApplicationModel] = self.get_unique_applications(samples=samples)
        return CaseModel(
            name=case.name,
            id=case.internal_id,
            data_analysis=self.get_case_analysis_data(case=case, analysis=analysis),
            samples=samples,
            applications=unique_applications,
        )

    def get_samples_data(self, case: Case, analysis_metadata: AnalysisModel) -> list[SampleModel]:
        """Extracts all the samples associated to a specific case and their attributes."""
        samples = list()
        case_samples: list[CaseSample] = self.status_db.get_case_samples_by_case_id(
            case_internal_id=case.internal_id
        )
        for case_sample in case_samples:
            sample: Sample = case_sample.sample
            lims_sample: dict[str, Any] = self.lims_api.sample(sample.internal_id)
            delivered_files: list[File] | None = (
                self.delivery_api.get_analysis_sample_delivery_files_by_sample(
                    case=case, sample=sample
                )
                if self.delivery_api.is_analysis_delivery(case.data_delivery)
                else None
            )
            delivered_fastq_files: list[File] | None = (
                self.delivery_api.get_fastq_delivery_files_by_sample(case=case, sample=sample)
                if self.delivery_api.is_fastq_delivery(case.data_delivery)
                else None
            )
            samples.append(
                SampleModel(
                    name=sample.name,
                    id=sample.internal_id,
                    ticket=sample.original_ticket,
                    sex=sample.sex,
                    source=lims_sample.get("source"),
                    tumour=sample.is_tumour,
                    application=self.get_sample_application(sample=sample, lims_sample=lims_sample),
                    methods=self.get_sample_methods_data(sample_id=sample.internal_id),
                    status=case_sample.status,
                    metadata=self.get_sample_metadata(
                        case=case, sample=sample, analysis_metadata=analysis_metadata
                    ),
                    timestamps=self.get_sample_timestamp_data(sample=sample),
                    delivered_files=delivered_files,
                    delivered_fastq_files=delivered_fastq_files,
                )
            )
        return samples

    def get_workflow_accreditation_limitation(self, application_tag: str) -> str | None:
        """Return workflow specific limitations given an application tag."""
        application_limitation: ApplicationLimitations = (
            self.status_db.get_application_limitation_by_tag_and_workflow(
                tag=application_tag, workflow=self.analysis_api.workflow
            )
        )
        return application_limitation.limitations if application_limitation else None

    def get_sample_application(
        self, sample: Sample, lims_sample: dict[str, Any]
    ) -> ApplicationModel:
        """Return the analysis application attributes for a sample."""
        application: Application = self.status_db.get_application_by_tag(
            tag=lims_sample.get("application")
        )
        return (
            ApplicationModel(
                tag=application.tag,
                version=sample.application_version.version,
                prep_category=application.prep_category,
                description=application.description,
                details=application.details,
                limitations=application.limitations,
                workflow_limitations=self.get_workflow_accreditation_limitation(application.tag),
                accredited=application.is_accredited,
                external=application.is_external,
            )
            if application
            else ApplicationModel()
        )

    @staticmethod
    def get_unique_applications(samples: list[SampleModel]) -> list[ApplicationModel]:
        """Returns the unique case associated applications."""
        applications = list()
        for sample in samples:
            if sample.application not in applications:
                applications.append(sample.application)
        return applications

    def get_sample_methods_data(self, sample_id: str) -> MethodsModel:
        """Fetches sample library preparation and sequencing methods from LIMS."""
        prep_method: str | None = self.lims_api.get_prep_method(lims_id=sample_id)
        sequencing_method: str | None = self.lims_api.get_sequencing_method(lims_id=sample_id)
        return MethodsModel(library_prep=prep_method, sequencing=sequencing_method)

    def get_case_analysis_data(self, case: Case, analysis: Analysis) -> DataAnalysisModel:
        """Return workflow attributes used for data analysis."""
        delivered_files: list[File] | None = (
            self.delivery_api.get_analysis_case_delivery_files(case)
            if self.delivery_api.is_analysis_delivery(case.data_delivery)
            else None
        )
        return DataAnalysisModel(
            comment=analysis.comment,
            customer_workflow=case.data_analysis,
            data_delivery=case.data_delivery,
            delivered_files=delivered_files,
            genome_build=self.analysis_api.get_genome_build(case.internal_id),
            panels=case.panels,
            pons=self.analysis_api.get_pons(case.internal_id),
            scout_files=self.get_scout_variants_files(case.internal_id),
            type=self.analysis_api.get_data_analysis_type(case.internal_id),
            variant_callers=self.analysis_api.get_variant_callers(case.internal_id),
            workflow=analysis.workflow,
            workflow_version=analysis.workflow_version,
        )

    def get_scout_variants_files(self, case_id: str) -> ScoutVariantsFiles:
        """Return files that will be uploaded to Scout."""
        return ScoutVariantsFiles()

    @staticmethod
    def get_sample_timestamp_data(sample: Sample) -> TimestampModel:
        """Return sample processing dates."""
        return TimestampModel(
            ordered_at=sample.ordered_at,
            received_at=sample.received_at,
            prepared_at=sample.prepared_at,
            reads_updated_at=sample.last_sequenced_at,
        )

    def get_sample_metadata(
        self,
        case: Case,
        sample: Sample,
        analysis_metadata: AnalysisModel,
    ) -> SampleMetadataModel:
        """Return sample metadata to include in the report."""
        raise NotImplementedError

    def is_report_accredited(
        self, samples: list[SampleModel], analysis_metadata: AnalysisModel
    ) -> bool:
        """Return whether the delivery report is accredited."""
        raise NotImplementedError

    def get_required_fields(self, case: CaseModel) -> dict:
        """Return dictionary with the delivery report required fields."""
        raise NotImplementedError

    @staticmethod
    def get_application_required_fields(case: CaseModel, required_fields: list) -> dict:
        """Return sample required fields."""
        required_sample_fields = dict()
        for application in case.applications:
            required_sample_fields.update({application.tag: required_fields})
        return required_sample_fields

    @staticmethod
    def get_sample_required_fields(case: CaseModel, required_fields: list) -> dict:
        """Return sample required fields."""
        required_sample_fields = dict()
        for sample in case.samples:
            required_sample_fields.update({sample.id: required_fields})
        return required_sample_fields

    @staticmethod
    def get_timestamp_required_fields(case: CaseModel, required_fields: list) -> dict:
        """Return sample timestamps required fields."""
        required_fields_copy: list = required_fields.copy()
        for sample in case.samples:
            if sample.application.external:
                required_fields_copy.remove("received_at")
                break
        return DeliveryReportAPI.get_sample_required_fields(
            case=case, required_fields=required_fields_copy
        )

    def get_hk_scout_file_tags(self, scout_key: ScoutUploadKey) -> list | None:
        """Return workflow specific uploaded to Scout Housekeeper file tags given a Scout key."""
        tags = self.analysis_api.get_scout_upload_case_tags().get(scout_key.value)
        return list(tags) if tags else None



"""Fixtures for the workflow tests."""

import datetime
import shutil
from pathlib import Path

import pytest

from cg.constants import SequencingFileTag
from cg.constants.constants import CaseActions, MicrosaltAppTags, MicrosaltQC, Workflow
from cg.meta.compress.compress import CompressAPI
from cg.meta.workflow.microsalt import MicrosaltAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.compression_data import CompressionData
from cg.models.orders.sample_base import ControlEnum
from cg.store.models import Case, Sample
from tests.cli.workflow.balsamic.conftest import (
    balsamic_housekeeper_dir,
    fastq_file_l_1_r_1,
    fastq_file_l_2_r_1,
    fastq_file_l_2_r_2,
)
from tests.meta.compress.conftest import compress_api, real_crunchy_api
from tests.mocks.balsamic_analysis_mock import MockBalsamicAnalysis
from tests.mocks.tb_mock import MockTB
from tests.store_helpers import StoreHelpers


@pytest.fixture(scope="function")
def analysis_api_balsamic(cg_context: CGConfig) -> MockBalsamicAnalysis:
    """BALSAMIC ReportAPI fixture."""
    return MockBalsamicAnalysis(cg_context)


@pytest.fixture(scope="function")
def populated_compress_spring_api(
    compress_api: CompressAPI, only_spring_bundle: dict, helpers
) -> CompressAPI:
    """Populated compress api fixture with only spring compressed fastq."""
    helpers.ensure_hk_bundle(compress_api.hk_api, only_spring_bundle)

    return compress_api


@pytest.fixture(scope="function")
def populated_compress_api_fastq_spring(
    compress_api: CompressAPI, spring_fastq_mix: dict, helpers
) -> CompressAPI:
    """Populated compress api fixture with both spring and fastq."""
    helpers.ensure_hk_bundle(compress_api.hk_api, spring_fastq_mix)

    return compress_api


@pytest.fixture(name="only_spring_bundle")
def only_spring_bundle() -> dict:
    """Return a dictionary with bundle info in the correct format."""
    return {
        "name": "ADM1",
        "created": "2019-12-24",
        "files": [
            {
                "path": "/path/HVCHCCCXY-l4t21_535422_S4_L004.spring",
                "archive": False,
                "tags": ["spring"],
            },
        ],
    }


@pytest.fixture(name="spring_fastq_mix")
def spring_fastq_mix(compression_object: CompressionData) -> dict:
    """Return a dictionary with bundle info including both fastq and spring files."""

    return {
        "name": "ADM1",
        "created": "2019-12-24",
        "files": [
            {
                "path": str(compression_object.spring_path),
                "archive": False,
                "tags": [SequencingFileTag.SPRING],
            },
            {
                "path": str(compression_object.fastq_first),
                "archive": False,
                "tags": [SequencingFileTag.FASTQ],
            },
            {
                "path": str(compression_object.fastq_second),
                "archive": False,
                "tags": [SequencingFileTag.FASTQ],
            },
        ],
    }


@pytest.fixture(name="microsalt_qc_pass_run_dir_path")
def microsalt_qc_pass_run_dir_path(
    microsalt_qc_pass_lims_project: str, microsalt_analysis_dir: Path
) -> Path:
    """Return a microsalt run dir path fixture that passes QC."""
    return Path(microsalt_analysis_dir, microsalt_qc_pass_lims_project)


@pytest.fixture(name="microsalt_qc_fail_run_dir_path")
def microsalt_qc_fail_run_dir_path(
    microsalt_qc_fail_lims_project: str, microsalt_analysis_dir: Path
) -> Path:
    """Return a microsalt run dir path fixture that fails QC."""
    return Path(microsalt_analysis_dir, microsalt_qc_fail_lims_project)


@pytest.fixture(name="microsalt_qc_pass_lims_project")
def microsalt_qc_pass_lims_project() -> str:
    """Return a microsalt LIMS project id that passes QC."""
    return "ACC22222_qc_pass"


@pytest.fixture(name="microsalt_qc_fail_lims_project")
def microsalt_qc_fail_lims_project() -> str:
    """Return a microsalt LIMS project id that fails QC."""
    return "ACC11111_qc_fail"


@pytest.fixture
def metrics_file_failing_qc(
    microsalt_qc_fail_run_dir_path: Path,
    microsalt_qc_fail_lims_project: str,
    tmp_path: Path,
) -> Path:
    """Return a metrics file that fails QC with corresponding samples in the database."""
    metrics_path = Path(microsalt_qc_fail_run_dir_path, f"{microsalt_qc_fail_lims_project}.json")
    temp_metrics_path = Path(tmp_path, metrics_path.name)
    shutil.copy(metrics_path, temp_metrics_path)
    return temp_metrics_path


@pytest.fixture
def metrics_file_passing_qc(
    microsalt_qc_pass_run_dir_path: Path,
    microsalt_qc_pass_lims_project: str,
    tmp_path: Path,
) -> Path:
    """Return a metrics file that pass QC with corresponding samples in the database."""
    metrics_path = Path(microsalt_qc_pass_run_dir_path, f"{microsalt_qc_pass_lims_project}.json")
    temp_metrics_path = Path(tmp_path, metrics_path.name)
    shutil.copy(metrics_path, temp_metrics_path)
    return temp_metrics_path


@pytest.fixture
def microsalt_metrics_file(
    microsalt_qc_fail_run_dir_path: Path, microsalt_qc_fail_lims_project: str
) -> Path:
    return Path(microsalt_qc_fail_run_dir_path, f"{microsalt_qc_fail_lims_project}.json")


@pytest.fixture(name="microsalt_case_qc_pass")
def microsalt_case_qc_pass() -> str:
    """Return a microsalt case to pass QC."""
    return "microsalt_case_qc_pass"


@pytest.fixture(name="microsalt_case_qc_fail")
def microsalt_case_qc_fail() -> str:
    """Return a microsalt case to fail QC."""
    return "microsalt_case_qc_fail"


@pytest.fixture(name="qc_pass_microsalt_samples")
def qc_pass_microsalt_samples() -> list[str]:
    """Return a list of 20 microsalt samples internal_ids."""
    return [f"ACC22222A{i}" for i in range(1, 21)]


@pytest.fixture(name="qc_fail_microsalt_samples")
def qc_fail_microsalt_samples() -> list[str]:
    """Return a list of 20 microsalt samples internal_ids."""
    return [f"ACC11111A{i}" for i in range(1, 21)]


@pytest.fixture(name="qc_microsalt_context")
def qc_microsalt_context(
    cg_context: CGConfig,
    helpers: StoreHelpers,
    microsalt_case_qc_pass: str,
    microsalt_case_qc_fail: str,
    qc_pass_microsalt_samples: list[str],
    qc_fail_microsalt_samples: list[str],
) -> CGConfig:
    """Return a Microsalt CG context."""
    cg_context.trailblazer_api_ = MockTB()
    analysis_api = MicrosaltAnalysisAPI(cg_context)
    store = analysis_api.status_db

    # Create MWR microsalt case that passes QC
    microsalt_case_qc_pass: Case = helpers.add_case(
        store=store,
        internal_id=microsalt_case_qc_pass,
        name=microsalt_case_qc_pass,
        data_analysis=Workflow.MICROSALT,
        action=CaseActions.RUNNING,
    )

    for sample in qc_pass_microsalt_samples[1:]:
        sample_to_add: Sample = helpers.add_sample(
            store=store,
            application_tag=MicrosaltAppTags.MWRNXTR003,
            application_type=MicrosaltAppTags.PREP_CATEGORY,
            internal_id=sample,
            reads=MicrosaltQC.TARGET_READS,
            last_sequenced_at=datetime.datetime.now(),
        )

        helpers.add_relationship(store=store, case=microsalt_case_qc_pass, sample=sample_to_add)

    # Add a negative control sample that passes the qc
    negative_control_sample: Sample = helpers.add_sample(
        store=store,
        internal_id=qc_pass_microsalt_samples[0],
        application_tag=MicrosaltAppTags.MWRNXTR003,
        application_type=MicrosaltAppTags.PREP_CATEGORY,
        reads=0,
        last_sequenced_at=datetime.datetime.now(),
        control=ControlEnum.negative,
    )
    helpers.add_relationship(
        store=store, case=microsalt_case_qc_pass, sample=negative_control_sample
    )

    # Create a microsalt MWX case that fails QC
    microsalt_case_qc_fail: Case = helpers.add_case(
        store=store,
        internal_id=microsalt_case_qc_fail,
        name=microsalt_case_qc_fail,
        data_analysis=Workflow.MICROSALT,
    )

    for sample in qc_fail_microsalt_samples:
        sample_to_add: Sample = helpers.add_sample(
            store=store,
            application_tag=MicrosaltAppTags.MWXNXTR003,
            application_type=MicrosaltAppTags.PREP_CATEGORY,
            internal_id=sample,
            reads=MicrosaltQC.TARGET_READS,
            last_sequenced_at=datetime.datetime.now(),
            control=ControlEnum.negative,
        )

        helpers.add_relationship(store=store, case=microsalt_case_qc_fail, sample=sample_to_add)

    # Setting the target reads to correspond with statusDB
    store.get_application_by_tag(tag=MicrosaltAppTags.MWRNXTR003).target_reads = (
        MicrosaltQC.TARGET_READS
    )
    store.get_application_by_tag(tag=MicrosaltAppTags.MWXNXTR003).target_reads = (
        MicrosaltQC.TARGET_READS
    )

    cg_context.meta_apis["analysis_api"] = analysis_api

    return cg_context


@pytest.fixture(name="rnafusion_metrics")
def rnafusion_metrics() -> dict[str, float]:
    """Return Rnafusion raw analysis metrics dictionary."""
    return {
        "after_filtering_gc_content": 0.516984,
        "after_filtering_q20_rate": 0.974834,
        "after_filtering_q30_rate": 0.929476,
        "after_filtering_read1_mean_length": 99.0,
        "before_filtering_total_reads": 149984042.0,
        "median_5prime_to_3prime_bias": 1.1211,
        "pct_adapter": 12.005654574904709,
        "pct_mrna_bases": 85.9731,
        "pct_ribosomal_bases": 0.6581,
        "pct_surviving": 99.42004630065911,
        "pct_duplication": 14.8643,
        "read_pairs_examined": 72391566.0,
        "uniquely_mapped_percent": 91.02,
    }


@pytest.fixture(name="mip_analysis_api")
def fixture_mip_analysis_api(
    cg_context: CGConfig, mip_hk_store, store_with_illumina_sequencing_data
) -> MipDNAAnalysisAPI:
    """Return a MIP analysis API."""
    analysis_api = MipDNAAnalysisAPI(cg_context)
    analysis_api.housekeeper_api = mip_hk_store
    analysis_api.status_db = store_with_illumina_sequencing_data
    return analysis_api


@pytest.fixture
def taxprofiler_metrics() -> dict[str, float]:
    """Return Taxprofiler raw analysis metrics dictionary."""
    return {
        "filtering_result_passed_filter_reads": 24810472.0,
        "reads_mapped": 19014950.0,
        "total_reads": 12400055,
        "paired_aligned_none": 1409340,
    }



from cg.models.orders.order import OrderIn
from cg.services.orders.store_order_services.store_pool_order import StorePoolOrderService
from cg.services.orders.submitters.order_submitter import OrderSubmitter
from cg.services.orders.validate_order_services.validate_pool_order import ValidatePoolOrderService


class PoolOrderSubmitter(OrderSubmitter):
    """
    Class for submitting pool orders.
    This class is used to submit orders for the following workflows:
    - Fluffy
    - RML (Ready made libraries)

    """

    def __init__(
        self,
        order_validation_service: ValidatePoolOrderService,
        order_store_service: StorePoolOrderService,
    ):
        self.order_validation_service = order_validation_service
        self.order_store_service = order_store_service

    def submit_order(self, order_in: OrderIn) -> dict:
        self.order_validation_service.validate_order(order_in)
        return self.order_store_service.store_order(order_in)



"""CLI for deleting records in statusDB """

import logging

import click

from cg.cli.delete.case import delete_case
from cg.cli.delete.cases import delete_cases
from cg.cli.delete.illumina_sequencing_run import delete_illumina_run
from cg.cli.delete.observations import (
    delete_available_observations,
    delete_observations,
)
from cg.cli.utils import CLICK_CONTEXT_SETTINGS

LOG = logging.getLogger(__name__)


@click.group(context_settings=CLICK_CONTEXT_SETTINGS)
def delete():
    """delete database records in CG."""
    LOG.info("Running CG delete")


delete.add_command(delete_case)
delete.add_command(delete_cases)
delete.add_command(delete_observations)
delete.add_command(delete_available_observations)
delete.add_command(delete_illumina_run)



import logging
import os
from pathlib import Path

from cg.services.deliver_files.file_fetcher.models import CaseFile, DeliveryFiles, SampleFile
from cg.services.deliver_files.file_formatter.abstract import DeliveryFileFormattingService
from cg.services.deliver_files.file_formatter.models import FormattedFile, FormattedFiles
from cg.services.deliver_files.file_formatter.utils.case_service import CaseFileFormatter
from cg.services.deliver_files.file_formatter.utils.sample_concatenation_service import (
    SampleFileConcatenationFormatter,
)
from cg.services.deliver_files.file_formatter.utils.sample_service import SampleFileFormatter

LOG = logging.getLogger(__name__)


class DeliveryFileFormatter(DeliveryFileFormattingService):
    """
    Format the files to be delivered in the generic format.
    Expected structure:
    <customer>/inbox/<ticket_id>/<case_name>/<case_files>
    <customer>/inbox/<ticket_id>/<sample_name>/<sample_files>
    """

    def __init__(
        self,
        case_file_formatter: CaseFileFormatter,
        sample_file_formatter: SampleFileFormatter | SampleFileConcatenationFormatter,
    ):
        self.case_file_formatter = case_file_formatter
        self.sample_file_formatter = sample_file_formatter

    def format_files(self, delivery_files: DeliveryFiles) -> FormattedFiles:
        """Format the files to be delivered and return the formatted files in the generic format."""
        LOG.debug("[FORMAT SERVICE] Formatting files for delivery")
        ticket_dir_path: Path = delivery_files.delivery_data.customer_ticket_inbox
        self._create_ticket_dir(ticket_dir_path)
        formatted_files: list[FormattedFile] = self._format_sample_and_case_files(
            sample_files=delivery_files.sample_files,
            case_files=delivery_files.case_files,
            ticket_dir_path=ticket_dir_path,
        )
        return FormattedFiles(files=formatted_files)

    def _format_sample_and_case_files(
        self, sample_files: list[SampleFile], case_files: list[CaseFile], ticket_dir_path: Path
    ) -> list[FormattedFile]:
        """Helper method to format both sample and case files."""
        formatted_files: list[FormattedFile] = self.sample_file_formatter.format_files(
            moved_files=sample_files,
            ticket_dir_path=ticket_dir_path,
        )
        if case_files:
            formatted_case_files: list[FormattedFile] = self.case_file_formatter.format_files(
                moved_files=case_files,
                ticket_dir_path=ticket_dir_path,
            )
            formatted_files.extend(formatted_case_files)
        return formatted_files

    @staticmethod
    def _create_ticket_dir(ticket_dir_path: Path) -> None:
        """Create the ticket directory if it does not exist."""
        os.makedirs(ticket_dir_path, exist_ok=True)



"""Module for demultiplex fixtures returning Path objects."""

import shutil
from pathlib import Path
from typing import List

import pytest

from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.nanopore_files import NanoporeDirsAndFiles

CORRECT_SAMPLE_SHEET: str = "CorrectSampleSheet.csv"


# Directory fixtures


@pytest.fixture(scope="session")
def demultiplex_fixtures(apps_dir: Path) -> Path:
    """Return the path to the demultiplex fixture directory."""
    return Path(apps_dir, "demultiplexing")


@pytest.fixture(scope="session")
def illumina_sequencing_runs_directory(demultiplex_fixtures: Path) -> Path:
    """Return the path to the sequenced flow cells fixture directory."""
    return Path(demultiplex_fixtures, DemultiplexingDirsAndFiles.SEQUENCING_RUNS_DIRECTORY_NAME)


@pytest.fixture(scope="session")
def illumina_demultiplexed_runs_directory(demultiplex_fixtures: Path) -> Path:
    """Return the path to the demultiplexed flow cells fixture directory."""
    return Path(demultiplex_fixtures, "demultiplexed-runs")


@pytest.fixture(scope="session")
def illumina_demux_all_directory(demultiplex_fixtures: Path) -> Path:
    """Return the path to the sequenced flow cells fixture directory."""
    return Path(demultiplex_fixtures, "sequencing_runs_demux_all")


@pytest.fixture(scope="session")
def nanopore_flow_cells_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the sequenced flow cells fixture directory."""
    return Path(demultiplex_fixtures, NanoporeDirsAndFiles.DATA_DIRECTORY)


@pytest.fixture(scope="session")
def run_parameters_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the run parameters fixture directory."""
    return Path(demultiplex_fixtures, "run_parameters")


@pytest.fixture(scope="session")
def sample_sheet_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the sample sheet fixture directory."""
    return Path(demultiplex_fixtures, "sample_sheets")


@pytest.fixture
def fastq_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the fastq files dir."""
    return Path(demultiplex_fixtures, "fastq")


@pytest.fixture
def spring_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the fastq files dir."""
    return Path(demultiplex_fixtures, "spring")


@pytest.fixture(scope="session")
def broken_flow_cells_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the broken or incomplete flow cells fixture directory."""
    return Path(demultiplex_fixtures, "sequencing_runs_broken")


@pytest.fixture(scope="session")
def illumina_demux_results_not_finished_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to a dir with demultiplexing results where nothing has been cleaned."""
    return Path(demultiplex_fixtures, "demultiplexed-runs-unfinished")


# Temporary directory fixtures


@pytest.fixture
def tmp_illumina_sequencing_runs_directory(
    tmp_path: Path, illumina_sequencing_runs_directory: Path
) -> Path:
    """
    Return the path to a temporary flow cells directory with flow cells ready for demultiplexing.
    Generates a copy of the original flow cells directory
    """
    original_dir = illumina_sequencing_runs_directory
    tmp_dir = Path(tmp_path, "sequencing-runs")

    return Path(shutil.copytree(original_dir, tmp_dir))


@pytest.fixture
def tmp_illumina_demultiplexed_runs_directory(tmp_path: Path) -> Path:
    """Return the path to a demultiplexed flow cell run directory."""
    demultiplexed_runs = Path(tmp_path, "demultiplexed-runs")
    demultiplexed_runs.mkdir(exist_ok=True)
    return demultiplexed_runs


@pytest.fixture
def tmp_illumina_demultiplexed_flow_cells_directory(
    tmp_path: Path, illumina_demultiplexed_runs_directory
) -> Path:
    """Return the path to a temporary demultiplex-runs directory.
    Generates a copy of the original demultiplexed-runs
    """
    original_dir = illumina_demultiplexed_runs_directory
    tmp_dir = Path(tmp_path, "demultiplexed-runs")
    return Path(shutil.copytree(original_dir, tmp_dir))


@pytest.fixture
def tmp_illumina_flow_cells_demux_all_directory(
    tmp_path: Path, illumina_demux_all_directory: Path
) -> Path:
    """
    Return the path to a temporary flow cells directory with flow cells ready for demultiplexing.
    Generates a copy of the original flow cells directory.
    This fixture is used for testing of the cg demultiplex all cmd.
    """
    original_dir = illumina_demux_all_directory
    tmp_dir = Path(tmp_path, "sequencing_runs_demux_all")

    return Path(shutil.copytree(original_dir, tmp_dir))


@pytest.fixture
def tmp_broken_flow_cells_directory(tmp_path: Path, broken_flow_cells_dir: Path) -> Path:
    """
    Return the path to a temporary flow cells directory with incomplete or broken flow cells.
    Generates a copy of the original flow cells directory
    """
    original_dir = broken_flow_cells_dir
    tmp_dir = Path(tmp_path, "broken_flow_cells")

    return Path(shutil.copytree(original_dir, tmp_dir))


# Canonical flow cell paths


@pytest.fixture(scope="session")
def hiseq_x_single_index_flow_cell_dir(
    illumina_sequencing_runs_directory, hiseq_x_single_index_flow_cell_name: str
) -> Path:
    """Return the path to a HiSeqX flow cell."""
    return Path(illumina_sequencing_runs_directory, hiseq_x_single_index_flow_cell_name)


@pytest.fixture(scope="session")
def hiseq_x_dual_index_flow_cell_dir(
    illumina_sequencing_runs_directory, hiseq_x_dual_index_flow_cell_name: str
) -> Path:
    """Return the path to a dual-index HiSeqX flow cell."""
    return Path(illumina_sequencing_runs_directory, hiseq_x_dual_index_flow_cell_name)


@pytest.fixture(scope="session")
def hiseq_2500_dual_index_flow_cell_dir(
    illumina_sequencing_runs_directory, hiseq_2500_dual_index_flow_cell_name: str
) -> Path:
    """Return the path to a HiSeq2500 flow cell."""
    return Path(illumina_sequencing_runs_directory, hiseq_2500_dual_index_flow_cell_name)


@pytest.fixture(scope="session")
def hiseq_2500_custom_index_flow_cell_dir(
    illumina_sequencing_runs_directory, hiseq_2500_custom_index_flow_cell_name: str
) -> Path:
    """Return the path to a HiSeq2500 flow cell."""
    return Path(illumina_sequencing_runs_directory, hiseq_2500_custom_index_flow_cell_name)


@pytest.fixture
def novaseq_6000_pre_1_5_kits_flow_cell_path(
    tmp_illumina_sequencing_runs_directory: Path, novaseq_6000_pre_1_5_kits_flow_cell_full_name: str
) -> Path:
    return Path(
        tmp_illumina_sequencing_runs_directory, novaseq_6000_pre_1_5_kits_flow_cell_full_name
    )


@pytest.fixture
def novaseq_6000_post_1_5_kits_flow_cell_path(
    tmp_illumina_sequencing_runs_directory: Path,
    novaseq_6000_post_1_5_kits_flow_cell_full_name: str,
) -> Path:
    return Path(
        tmp_illumina_sequencing_runs_directory, novaseq_6000_post_1_5_kits_flow_cell_full_name
    )


@pytest.fixture
def novaseq_x_flow_cell_dir(
    illumina_sequencing_runs_directory: Path, novaseq_x_flow_cell_full_name: str
) -> Path:
    """Return the path to a NovaSeqX flow cell."""
    return Path(illumina_sequencing_runs_directory, novaseq_x_flow_cell_full_name)


# Tmp flow cell paths


@pytest.fixture
def tmp_novaseq_6000_pre_1_5_kits_flow_cell_path(
    tmp_illumina_sequencing_runs_directory: Path, novaseq_6000_pre_1_5_kits_flow_cell_full_name: str
) -> Path:
    return Path(
        tmp_illumina_sequencing_runs_directory, novaseq_6000_pre_1_5_kits_flow_cell_full_name
    )


@pytest.fixture
def tmp_illumina_flow_cells_demux_results_not_finished_directory(
    tmp_path: Path, illumina_demux_results_not_finished_dir: Path
) -> Path:
    """Return the path to a temporary flow cells directory with unfinished demultiplexing results."""
    original_dir: Path = illumina_demux_results_not_finished_dir
    tmp_dir = Path(tmp_path, "demultiplexed-runs-unfinished")

    return Path(shutil.copytree(original_dir, tmp_dir))


@pytest.fixture
def tmp_flow_cell_directory_bcl_convert(tmp_illumina_sequencing_runs_directory: Path) -> Path:
    """Return a path to a flow cell directory with the run parameters present."""
    return Path(tmp_illumina_sequencing_runs_directory, "211101_A00187_0615_AHLG5GDRZZ")


@pytest.fixture
def tmp_flow_cell_with_bcl2fastq_sample_sheet(
    tmp_broken_flow_cells_directory: Path, hiseq_x_single_index_flow_cell_name: str
) -> Path:
    """Return a path to a flow cell directory with a BCL2FASTQ sample sheet for translation."""
    return Path(tmp_broken_flow_cells_directory, hiseq_x_single_index_flow_cell_name)


@pytest.fixture
def tmp_flow_cell_without_run_parameters_path(
    tmp_broken_flow_cells_directory: Path, hiseq_2500_custom_index_flow_cell_name: str
) -> Path:
    """Return a path to a flow cell directory with the run parameters missing."""
    return Path(tmp_broken_flow_cells_directory, hiseq_2500_custom_index_flow_cell_name)


@pytest.fixture
def tmp_novaseq_6000_pre_1_5_kits_flow_cell_without_sample_sheet_path(
    tmp_broken_flow_cells_directory: Path,
    novaseq_6000_pre_1_5_kits_flow_cell_full_name: str,
) -> Path:
    """Return a path to a flow cell directory with the sample sheet missing."""
    return Path(tmp_broken_flow_cells_directory, novaseq_6000_pre_1_5_kits_flow_cell_full_name)


@pytest.fixture
def tmp_novaseq_6000_post_1_5_kits_flow_cell_without_sample_sheet_path(
    tmp_broken_flow_cells_directory: Path,
    novaseq_6000_post_1_5_kits_flow_cell_full_name: str,
) -> Path:
    """Return a path to a flow cell directory with the sample sheet missing."""
    return Path(tmp_broken_flow_cells_directory, novaseq_6000_post_1_5_kits_flow_cell_full_name)


@pytest.fixture
def tmp_novaseq_x_without_sample_sheet_flow_cell_path(
    tmp_broken_flow_cells_directory: Path,
    novaseq_x_flow_cell_full_name: str,
) -> Path:
    """Return a path to a flow cell directory with the sample sheet missing."""
    return Path(tmp_broken_flow_cells_directory, novaseq_x_flow_cell_full_name)


@pytest.fixture
def tmp_novaseqx_flow_cell_directory(tmp_path: Path, novaseq_x_flow_cell_full_name: str) -> Path:
    """Return the path to a NovaseqX flow cell directory."""
    return Path(tmp_path, novaseq_x_flow_cell_full_name)


@pytest.fixture
def tmp_flow_cells_directory_ready_for_demultiplexing(
    hiseq_2500_dual_index_flow_cell_name: str, tmp_illumina_sequencing_runs_directory
) -> Path:
    """Return a path to a flow cell directory with the run parameters missing."""
    return Path(tmp_illumina_sequencing_runs_directory, hiseq_2500_dual_index_flow_cell_name)


# Temporary demultiplexed runs fixtures


@pytest.fixture
def tmp_demultiplexed_flow_cell_no_fastq_files(
    tmp_illumina_flow_cells_demux_results_not_finished_directory: Path,
    novaseq_6000_post_1_5_kits_flow_cell_full_name: str,
) -> Path:
    """Return the path to a demultiplexed flow cell directory without fastq files."""
    return Path(
        tmp_illumina_flow_cells_demux_results_not_finished_directory,
        novaseq_6000_post_1_5_kits_flow_cell_full_name,
    )


@pytest.fixture
def tmp_demultiplexed_novaseq_6000_post_1_5_kits_path(
    tmp_illumina_demultiplexed_flow_cells_directory: Path,
    novaseq_6000_post_1_5_kits_flow_cell_full_name: str,
) -> Path:
    """Return the path to a demultiplexed flow cell directory without fastq files."""
    return Path(
        tmp_illumina_demultiplexed_flow_cells_directory,
        novaseq_6000_post_1_5_kits_flow_cell_full_name,
    )


# Fixtures for test demultiplex flow cell
@pytest.fixture
def tmp_empty_demultiplexed_runs_directory(
    tmp_illumina_demultiplexed_flow_cells_directory,
) -> Path:
    return Path(tmp_illumina_demultiplexed_flow_cells_directory, "empty")


@pytest.fixture(scope="function")
def novaseqx_latest_analysis_version() -> str:
    """Return the latest analysis version for NovaseqX analysis data directory."""
    return "2"


def add_novaseqx_analysis_data(novaseqx_flow_cell_directory: Path, analysis_version: str):
    """Add NovaSeqX analysis data to a flow cell directory."""
    analysis_path: Path = Path(
        novaseqx_flow_cell_directory, DemultiplexingDirsAndFiles.ANALYSIS, analysis_version
    )
    analysis_path.mkdir(parents=True)
    analysis_path.joinpath(DemultiplexingDirsAndFiles.COPY_COMPLETE).touch()
    data = analysis_path.joinpath(DemultiplexingDirsAndFiles.DATA)
    data.mkdir()
    data.joinpath(DemultiplexingDirsAndFiles.ANALYSIS_COMPLETED).touch()


@pytest.fixture(scope="function")
def novaseqx_flow_cell_dir_with_analysis_data(
    tmp_novaseqx_flow_cell_directory: Path, novaseqx_latest_analysis_version: str
) -> Path:
    """Return the path to a NovaseqX flow cell directory with multiple analysis data directories."""
    add_novaseqx_analysis_data(tmp_novaseqx_flow_cell_directory, "0")
    add_novaseqx_analysis_data(tmp_novaseqx_flow_cell_directory, "1")
    add_novaseqx_analysis_data(tmp_novaseqx_flow_cell_directory, novaseqx_latest_analysis_version)
    return tmp_novaseqx_flow_cell_directory


@pytest.fixture(scope="function")
def post_processed_novaseqx_flow_cell(novaseqx_flow_cell_dir_with_analysis_data: Path) -> Path:
    """Return the path to a NovaseqX flow cell that is post processed."""
    Path(
        novaseqx_flow_cell_dir_with_analysis_data,
        DemultiplexingDirsAndFiles.QUEUED_FOR_POST_PROCESSING,
    ).touch()
    return novaseqx_flow_cell_dir_with_analysis_data


@pytest.fixture(scope="function")
def novaseqx_flow_cell_analysis_incomplete(
    tmp_novaseqx_flow_cell_directory: Path, novaseqx_latest_analysis_version: str
) -> Path:
    """
    Return the path to a flow cell for which the analysis is not complete.
    It misses the ANALYSIS_COMPLETED file.
    """
    Path(
        tmp_novaseqx_flow_cell_directory,
        DemultiplexingDirsAndFiles.ANALYSIS,
        novaseqx_latest_analysis_version,
    ).mkdir(parents=True)
    Path(
        tmp_novaseqx_flow_cell_directory,
        DemultiplexingDirsAndFiles.ANALYSIS,
        novaseqx_latest_analysis_version,
        DemultiplexingDirsAndFiles.COPY_COMPLETE,
    ).touch()
    return tmp_novaseqx_flow_cell_directory


@pytest.fixture(scope="function")
def demultiplex_not_complete_novaseqx_flow_cell(tmp_file: Path) -> Path:
    """Return the path to a NovaseqX flow cell for which demultiplexing is not complete."""
    return tmp_file


@pytest.fixture
def novaseq_x_demux_runs_dir(
    illumina_demultiplexed_runs_directory: Path, novaseq_x_flow_cell_full_name: str
) -> Path:
    """Return the path to a NovaSeqX flow cell."""
    return Path(illumina_demultiplexed_runs_directory, novaseq_x_flow_cell_full_name)


@pytest.fixture
def hiseq_2500_dual_index_demux_runs_dir(
    illumina_sequencing_runs_directory, hiseq_2500_dual_index_flow_cell_name: str
) -> Path:
    """Return the path to a HiSeq2500 flow cell."""
    return Path(illumina_sequencing_runs_directory, hiseq_2500_dual_index_flow_cell_name)


# Path to run parameter files


@pytest.fixture(scope="session")
def run_parameters_wrong_instrument(run_parameters_dir: Path) -> Path:
    """Return a NovaSeqX run parameters file path with a wrong instrument value."""
    return Path(run_parameters_dir, "RunParameters_novaseq_X_wrong_instrument.xml")


@pytest.fixture(scope="session")
def hiseq_x_single_index_run_parameters_path(
    hiseq_x_single_index_flow_cell_dir: Path,
) -> Path:
    """Return the path to a HiSeqX run parameters file with single index."""
    return Path(
        hiseq_x_single_index_flow_cell_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE
    )


@pytest.fixture(scope="session")
def hiseq_x_dual_index_run_parameters_path(
    hiseq_x_dual_index_flow_cell_dir: Path,
) -> Path:
    """Return the path to a HiSeqX run parameters file with dual index."""
    return Path(
        hiseq_x_dual_index_flow_cell_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE
    )


@pytest.fixture(scope="session")
def hiseq_2500_dual_index_run_parameters_path(
    hiseq_2500_dual_index_flow_cell_dir: Path,
) -> Path:
    """Return the path to a HiSeq2500 run parameters file with dual index."""
    return Path(
        hiseq_2500_dual_index_flow_cell_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE
    )


@pytest.fixture(scope="session")
def hiseq_2500_custom_index_run_parameters_path(
    hiseq_2500_custom_index_flow_cell_dir: Path,
) -> Path:
    """Return the path to a HiSeq2500 run parameters file with custom index."""
    return Path(
        hiseq_2500_custom_index_flow_cell_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE
    )


@pytest.fixture
def novaseq_6000_run_parameters_path(novaseq_6000_pre_1_5_kits_flow_cell_path: Path) -> Path:
    """Return the path to a NovaSeq6000 run parameters file."""
    return Path(
        novaseq_6000_pre_1_5_kits_flow_cell_path,
        DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE,
    )


@pytest.fixture
def novaseq_6000_run_parameters_pre_1_5_kits_path(
    novaseq_6000_pre_1_5_kits_flow_cell_path: Path,
) -> Path:
    """Return the path to a NovaSeq6000 pre 1.5 kit run parameters file."""
    return Path(
        novaseq_6000_pre_1_5_kits_flow_cell_path,
        DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE,
    )


@pytest.fixture
def novaseq_6000_run_parameters_post_1_5_kits_path(
    novaseq_6000_post_1_5_kits_flow_cell_path: Path,
) -> Path:
    """Return the path to a NovaSeq6000 post 1.5 kit run parameters file."""
    return Path(
        novaseq_6000_post_1_5_kits_flow_cell_path,
        DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE,
    )


@pytest.fixture
def novaseq_x_run_parameters_path(novaseq_x_flow_cell_dir: Path) -> Path:
    """Return the path to a NovaSeqX run parameters file."""
    return Path(novaseq_x_flow_cell_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE)


@pytest.fixture
def novaseq_x_run_parameters_name_node_path(run_parameters_dir: Path) -> Path:
    return Path(
        run_parameters_dir,
        "RunParameters_novaseq_x_flow_cell_name_node.xml",
    )


@pytest.fixture(scope="module")
def run_parameters_missing_versions_path(
    run_parameters_dir: Path,
) -> Path:
    """Return a NovaSeq6000 run parameters path without software and reagent kit versions."""
    return Path(run_parameters_dir, "RunParameters_novaseq_no_software_nor_reagent_version.xml")


# Path to sample sheets


@pytest.fixture
def hiseq_x_single_index_sample_sheet_path(hiseq_x_single_index_flow_cell_dir: Path) -> Path:
    """Return the path to a single-index HiSeqX sample sheet."""
    return Path(
        hiseq_x_single_index_flow_cell_dir, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
    )


@pytest.fixture
def hiseq_x_dual_index_sample_sheet_path(hiseq_x_dual_index_flow_cell_dir: Path) -> Path:
    """Return the path to a dual-index HiSeqX sample sheet."""
    return Path(hiseq_x_dual_index_flow_cell_dir, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME)


@pytest.fixture
def hiseq_2500_dual_index_sample_sheet_path(hiseq_2500_dual_index_flow_cell_dir: Path) -> Path:
    """Return the path to a dual-index HiSeq2500 sample sheet."""
    return Path(
        hiseq_2500_dual_index_flow_cell_dir, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
    )


@pytest.fixture
def hiseq_2500_custom_index_sample_sheet_path(hiseq_2500_custom_index_flow_cell_dir: Path) -> Path:
    """Return the path to a custom-index HiSeq2500 sample sheet."""
    return Path(
        hiseq_2500_custom_index_flow_cell_dir, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
    )


@pytest.fixture
def novaseq_6000_pre_1_5_kits_correct_sample_sheet_path(
    novaseq_6000_pre_1_5_kits_flow_cell_path: Path,
) -> Path:
    return Path(novaseq_6000_pre_1_5_kits_flow_cell_path, CORRECT_SAMPLE_SHEET)


@pytest.fixture
def novaseq_6000_post_1_5_kits_correct_sample_sheet_path(
    novaseq_6000_post_1_5_kits_flow_cell_path: Path,
) -> Path:
    return Path(novaseq_6000_post_1_5_kits_flow_cell_path, CORRECT_SAMPLE_SHEET)


@pytest.fixture
def novaseq_x_correct_sample_sheet(novaseq_x_flow_cell_dir: Path) -> Path:
    return Path(novaseq_x_flow_cell_dir, CORRECT_SAMPLE_SHEET)


@pytest.fixture
def seven_canonical_flow_cell_sample_sheets(
    hiseq_x_single_index_sample_sheet_path: Path,
    hiseq_x_dual_index_sample_sheet_path: Path,
    hiseq_2500_dual_index_sample_sheet_path: Path,
    hiseq_2500_custom_index_sample_sheet_path: Path,
    novaseq_6000_pre_1_5_kits_correct_sample_sheet_path: Path,
    novaseq_6000_post_1_5_kits_correct_sample_sheet_path: Path,
    novaseq_x_correct_sample_sheet: Path,
) -> list[Path]:
    """Return a dictionary with seven canonical flow cell sample sheets."""
    return [
        hiseq_x_single_index_sample_sheet_path,
        hiseq_x_dual_index_sample_sheet_path,
        hiseq_2500_dual_index_sample_sheet_path,
        hiseq_2500_custom_index_sample_sheet_path,
        novaseq_6000_pre_1_5_kits_correct_sample_sheet_path,
        novaseq_6000_post_1_5_kits_correct_sample_sheet_path,
        novaseq_x_correct_sample_sheet,
    ]


@pytest.fixture
def canonical_flow_cell_ids_and_sample_sheet_paths(
    seven_canonical_flow_cell_ids: list[str],
    seven_canonical_flow_cell_sample_sheets: list[Path],
) -> dict[str, Path]:
    """Return a dictionary with canonical flow cell ids as keys and sample sheet paths as values."""
    sample_sheet_paths: dict[str, Path] = {}
    for flow_cell_id, sample_sheet_path in zip(
        seven_canonical_flow_cell_ids, seven_canonical_flow_cell_sample_sheets
    ):
        sample_sheet_paths[flow_cell_id] = sample_sheet_path
    return sample_sheet_paths


@pytest.fixture
def novaseq_6000_sample_sheet_with_reversed_cycles(sample_sheet_dir: Path) -> Path:
    """Return the path to a NovaSeq6000 sample sheet with reversed index2 cycles."""
    return Path(sample_sheet_dir, "novaseq_6000_sample_sheet_with_reversed_cycles.csv")


@pytest.fixture
def novaseq_x_sample_sheet_with_forward_cycles(sample_sheet_dir: Path) -> Path:
    """Return the path to a NovaSeqX sample sheet with forward index2 cycles."""
    return Path(sample_sheet_dir, "novaseq_x_sample_sheet_with_forward_cycles.csv")


# Path to other flow cell attributes


@pytest.fixture
def novaseq_x_manifest_file(novaseq_x_flow_cell_dir: Path) -> Path:
    """Return the path to a NovaSeqX manifest file."""
    return Path(novaseq_x_flow_cell_dir, "Manifest.tsv")


@pytest.fixture(name="fastq_file")
def fastq_file(fastq_dir: Path) -> Path:
    """Return the path to a FASTQ file."""
    return Path(fastq_dir, "dummy_run_R1_001.fastq.gz")


@pytest.fixture(name="fastq_file_father")
def fastq_file_father(fastq_dir: Path) -> Path:
    """Return the path to a FASTQ file."""
    return Path(fastq_dir, "fastq_run_R1_001.fastq.gz")


@pytest.fixture(name="spring_file")
def spring_file(spring_dir: Path) -> Path:
    """Return the path to an existing spring file."""
    return Path(spring_dir, "dummy_run_001.spring")


@pytest.fixture(name="spring_meta_data_file")
def spring_meta_data_file(spring_dir: Path) -> Path:
    """Return the path to an existing spring file."""
    return Path(spring_dir, "dummy_spring_meta_data.json")


@pytest.fixture(name="spring_file_father")
def spring_file_father(spring_dir: Path) -> Path:
    """Return the path to a second existing spring file."""
    return Path(spring_dir, "dummy_run_002.spring")


# temporary fastq paths for canonical illumina runs
def make_tmp_fastq_paths(tmp_path, run_id, sample_ids: List[str]) -> dict[str, Path]:
    fastq_paths: dict[str, Path] = {}
    tmp_path_parent: Path = Path(tmp_path, run_id)
    tmp_path_parent.mkdir(exist_ok=True)
    for sample_id in sample_ids:
        fastq_path = Path(tmp_path_parent, f"{sample_id}_S1_L001_R1_001.fastq.gz")
        fastq_path.touch()
        fastq_paths[sample_id] = fastq_path
    return fastq_paths


@pytest.fixture
def tmp_hiseq_x_single_index_fastq_path(
    tmp_illumina_demultiplexed_runs_directory: Path,
    hiseq_x_single_index_flow_cell_name: str,
    selected_hiseq_x_single_index_sample_ids: list[str],
) -> dict[str, Path]:
    """Return the paths to a temporary single-index HiSeqX FASTQ file."""
    return make_tmp_fastq_paths(
        tmp_illumina_demultiplexed_runs_directory,
        hiseq_x_single_index_flow_cell_name,
        selected_hiseq_x_single_index_sample_ids,
    )


@pytest.fixture
def tmp_hiseq_x_dual_index_fastq_path(
    tmp_illumina_demultiplexed_runs_directory: Path,
    hiseq_x_dual_index_flow_cell_name: str,
    selected_hiseq_x_dual_index_sample_ids: list[str],
) -> dict[str, Path]:
    """Return the path to a temporary dual-index HiSeqX FASTQ file."""
    return make_tmp_fastq_paths(
        tmp_illumina_demultiplexed_runs_directory,
        hiseq_x_dual_index_flow_cell_name,
        selected_hiseq_x_dual_index_sample_ids,
    )


@pytest.fixture
def tmp_hiseq_2500_dual_index_fastq_path(
    tmp_illumina_demultiplexed_runs_directory: Path,
    hiseq_2500_dual_index_flow_cell_name: str,
    selected_hiseq_2500_dual_index_sample_ids: list[str],
) -> dict[str, Path]:
    """Return the path to a temporary dual-index HiSeq2500 FASTQ file."""
    return make_tmp_fastq_paths(
        tmp_illumina_demultiplexed_runs_directory,
        hiseq_2500_dual_index_flow_cell_name,
        selected_hiseq_2500_dual_index_sample_ids,
    )


@pytest.fixture
def tmp_hiseq_2500_custom_index_fastq_path(
    tmp_illumina_demultiplexed_runs_directory: Path,
    hiseq_2500_custom_index_flow_cell_name: str,
    selected_hiseq_2500_custom_index_sample_ids: list[str],
) -> dict[str, Path]:
    """Return the path to a temporary custom-index HiSeq2500 FASTQ file."""
    return make_tmp_fastq_paths(
        tmp_illumina_demultiplexed_runs_directory,
        hiseq_2500_custom_index_flow_cell_name,
        selected_hiseq_2500_custom_index_sample_ids,
    )


@pytest.fixture
def tmp_novaseq_6000_pre_1_5_kits_fastq_path(
    tmp_illumina_demultiplexed_runs_directory: Path,
    novaseq_6000_pre_1_5_kits_flow_cell_full_name: str,
    selected_novaseq_6000_pre_1_5_kits_sample_ids: list[str],
) -> dict[str, Path]:
    """Return the path to a temporary NovaSeq6000 pre 1.5 kits FASTQ file."""
    return make_tmp_fastq_paths(
        tmp_illumina_demultiplexed_runs_directory,
        novaseq_6000_pre_1_5_kits_flow_cell_full_name,
        selected_novaseq_6000_pre_1_5_kits_sample_ids,
    )


@pytest.fixture
def tmp_novaseq_6000_post_1_5_kits_fastq_path(
    tmp_illumina_demultiplexed_runs_directory: Path,
    novaseq_6000_post_1_5_kits_flow_cell_full_name: str,
    selected_novaseq_6000_post_1_5_kits_sample_ids: list[str],
) -> dict[str, Path]:
    """Return the path to a temporary NovaSeq6000 post 1.5 kits FASTQ file."""
    return make_tmp_fastq_paths(
        tmp_illumina_demultiplexed_runs_directory,
        novaseq_6000_post_1_5_kits_flow_cell_full_name,
        selected_novaseq_6000_post_1_5_kits_sample_ids,
    )


@pytest.fixture
def tmp_novaseq_x_fastq_path(
    tmp_illumina_demultiplexed_runs_directory: Path,
    novaseq_x_flow_cell_full_name: str,
    selected_novaseq_x_sample_ids: list[str],
) -> dict[str, Path]:
    """Return the path to a temporary NovaSeqX FASTQ file."""
    return make_tmp_fastq_paths(
        tmp_illumina_demultiplexed_runs_directory,
        novaseq_x_flow_cell_full_name,
        selected_novaseq_x_sample_ids,
    )


@pytest.fixture
def tmp_fastq_files_for_canonical_illumina_demultiplexed_runs(
    tmp_hiseq_x_single_index_fastq_path: dict[str, Path],
    tmp_hiseq_x_dual_index_fastq_path: dict[str, Path],
    tmp_hiseq_2500_dual_index_fastq_path: dict[str, Path],
    tmp_hiseq_2500_custom_index_fastq_path: dict[str, Path],
    tmp_novaseq_6000_pre_1_5_kits_fastq_path: dict[str, Path],
    tmp_novaseq_6000_post_1_5_kits_fastq_path: dict[str, Path],
    tmp_novaseq_x_fastq_path: dict[str, Path],
) -> list[dict[str, Path]]:
    return [
        tmp_hiseq_x_single_index_fastq_path,
        tmp_hiseq_x_dual_index_fastq_path,
        tmp_hiseq_2500_dual_index_fastq_path,
        tmp_hiseq_2500_custom_index_fastq_path,
        tmp_novaseq_6000_pre_1_5_kits_fastq_path,
        tmp_novaseq_6000_post_1_5_kits_fastq_path,
        tmp_novaseq_x_fastq_path,
    ]


@pytest.fixture
def tmp_fastq_files_for_all_canonical_illumina_demultiplexed_runs(
    seven_canonical_flow_cell_ids: list[str],
    tmp_fastq_files_for_canonical_illumina_demultiplexed_runs: list[dict[str, Path]],
) -> dict[str, dict[str, Path]]:
    return {
        flow_cell_id: fastq_paths
        for flow_cell_id, fastq_paths in zip(
            seven_canonical_flow_cell_ids, tmp_fastq_files_for_canonical_illumina_demultiplexed_runs
        )
    }



from cg.apps.orderform.orderform_parser import OrderformParser
from cg.constants import DataDelivery, Workflow
from cg.exc import OrderFormError
from cg.models.orders.json_sample import JsonSample
from cg.models.orders.order import OrderType


class JsonOrderformParser(OrderformParser):
    ACCEPTED_DATA_ANALYSES: list[str] = [
        Workflow.MIP_DNA,
        Workflow.MIP_RNA,
        Workflow.FLUFFY,
        Workflow.BALSAMIC,
    ]
    NO_VALUE: str = "no_value"
    samples: list[JsonSample] = []

    def get_project_type(self) -> str:
        """Determine the project type."""

        data_analyses: set[str] = {sample.data_analysis.lower() for sample in self.samples}

        if len(data_analyses) > 1:
            raise OrderFormError(f"mixed 'Data Analysis' types: {', '.join(data_analyses)}")

        data_analysis: str = data_analyses.pop()
        if data_analysis in self.ACCEPTED_DATA_ANALYSES:
            return data_analysis

        raise OrderFormError(f"Unsupported data_analysis: {data_analyses} for json data")

    @staticmethod
    def project_type_to_order_type(project_type: str) -> str:
        """In the case where data delivery was not defined we map from project type"""
        project_to_order = {
            OrderType.METAGENOME: DataDelivery.FASTQ,
            OrderType.FASTQ: DataDelivery.FASTQ,
            OrderType.RML: DataDelivery.FASTQ,
            OrderType.MIP_RNA: DataDelivery.ANALYSIS_FILES,
            OrderType.FLUFFY: DataDelivery.STATINA,
        }
        if project_type not in project_to_order:
            raise OrderFormError(f"Could not find data delivery for: {project_type}")
        return project_to_order[project_type]

    def get_data_delivery(self) -> str:
        """Determine the order_data delivery type."""

        data_deliveries = {sample.data_delivery for sample in self.samples}

        if len(data_deliveries) > 1:
            raise OrderFormError(f"mixed 'Data Delivery' types: {', '.join(data_deliveries)}")

        data_delivery = data_deliveries.pop()

        if data_delivery == self.NO_VALUE:
            return str(self.project_type_to_order_type(self.project_type))

        if data_delivery == str(DataDelivery.NIPT_VIEWER):
            return str(DataDelivery.STATINA)

        try:
            return str(DataDelivery(data_delivery))
        except ValueError:
            raise OrderFormError(f"Unsupported order_data delivery: {data_delivery}")

    def parse_orderform(self, order_data: dict) -> None:
        """Parse order form in JSON format."""

        self.samples: list[JsonSample] = [
            JsonSample.model_validate(sample_data) for sample_data in order_data.get("samples", [])
        ]
        if not self.samples:
            raise OrderFormError("orderform doesn't contain any samples")

        self.project_type = self.get_project_type()
        self.delivery_type = self.get_data_delivery()
        self.customer_id = order_data["customer"].lower()
        self.order_comment = order_data.get("comment")
        self.order_name = order_data.get("name")



from unittest.mock import MagicMock

import pytest
from _pytest.fixtures import FixtureRequest
from pydantic import BaseModel

from cg.constants import DataDelivery, Workflow
from cg.services.deliver_files.deliver_files_service.deliver_files_service import (
    DeliverFilesService,
)
from cg.services.deliver_files.deliver_files_service.deliver_files_service_factory import (
    DeliveryServiceFactory,
)
from cg.services.deliver_files.file_fetcher.abstract import FetchDeliveryFilesService
from cg.services.deliver_files.file_fetcher.analysis_raw_data_service import (
    RawDataAndAnalysisDeliveryFileFetcher,
)
from cg.services.deliver_files.file_fetcher.analysis_service import AnalysisDeliveryFileFetcher
from cg.services.deliver_files.file_fetcher.raw_data_service import RawDataDeliveryFileFetcher
from cg.services.deliver_files.file_formatter.utils.sample_concatenation_service import (
    SampleFileConcatenationFormatter,
)
from cg.services.deliver_files.file_formatter.utils.sample_service import SampleFileFormatter
from cg.services.deliver_files.file_mover.service import DeliveryFilesMover
from cg.services.deliver_files.tag_fetcher.abstract import FetchDeliveryFileTagsService
from cg.services.deliver_files.tag_fetcher.sample_and_case_service import (
    SampleAndCaseDeliveryTagsFetcher,
)


class DeliveryServiceScenario(BaseModel):
    """Model holding the necessary parameters to build a specific delivery service."""

    app_tag: str
    data_analysis: Workflow
    delivery_type: DataDelivery
    expected_tag_fetcher: type[FetchDeliveryFileTagsService]
    expected_file_fetcher: type[FetchDeliveryFilesService]
    expected_file_mover: type[DeliveryFilesMover]
    expected_sample_file_formatter: type[SampleFileFormatter | SampleFileConcatenationFormatter]
    store_name: str


@pytest.mark.parametrize(
    "scenario",
    [
        DeliveryServiceScenario(
            app_tag="MWRNXTR003",
            data_analysis=Workflow.RAW_DATA,
            delivery_type=DataDelivery.FASTQ,
            expected_tag_fetcher=SampleAndCaseDeliveryTagsFetcher,
            expected_file_fetcher=RawDataDeliveryFileFetcher,
            expected_file_mover=DeliveryFilesMover,
            expected_sample_file_formatter=SampleFileConcatenationFormatter,
            store_name="microbial_store",
        ),
        DeliveryServiceScenario(
            app_tag="VWGDPTR001",
            data_analysis=Workflow.MUTANT,
            delivery_type=DataDelivery.ANALYSIS_FILES,
            expected_tag_fetcher=SampleAndCaseDeliveryTagsFetcher,
            expected_file_fetcher=AnalysisDeliveryFileFetcher,
            expected_file_mover=DeliveryFilesMover,
            expected_sample_file_formatter=SampleFileFormatter,
            store_name="mutant_store",
        ),
        DeliveryServiceScenario(
            app_tag="PANKTTR020",
            data_analysis=Workflow.BALSAMIC,
            delivery_type=DataDelivery.FASTQ_ANALYSIS,
            expected_tag_fetcher=SampleAndCaseDeliveryTagsFetcher,
            expected_file_fetcher=RawDataAndAnalysisDeliveryFileFetcher,
            expected_file_mover=DeliveryFilesMover,
            expected_sample_file_formatter=SampleFileFormatter,
            store_name="applications_store",
        ),
    ],
    ids=["microbial-fastq", "SARS-COV2", "Targeted"],
)
def test_build_delivery_service(scenario: DeliveryServiceScenario, request: FixtureRequest):
    # GIVEN a delivery service builder with mocked store and hk_api
    builder = DeliveryServiceFactory(
        store=request.getfixturevalue(scenario.store_name),
        hk_api=MagicMock(),
        rsync_service=MagicMock(),
        tb_service=MagicMock(),
        analysis_service=MagicMock(),
    )

    # GIVEN a case with mocked sample app tag
    case_mock = MagicMock()
    case_mock.data_delivery = scenario.delivery_type
    case_mock.data_analysis = scenario.data_analysis
    case_mock.samples = [
        MagicMock(application_version=MagicMock(application=MagicMock(tag=scenario.app_tag)))
    ]

    # WHEN building a delivery service
    delivery_service: DeliverFilesService = builder.build_delivery_service(case=case_mock)

    # THEN the correct file formatter and file fetcher services are used
    assert isinstance(delivery_service.file_manager.tags_fetcher, scenario.expected_tag_fetcher)
    assert isinstance(delivery_service.file_manager, scenario.expected_file_fetcher)
    assert isinstance(delivery_service.file_mover, scenario.expected_file_mover)
    assert isinstance(
        delivery_service.file_formatter.sample_file_formatter,
        scenario.expected_sample_file_formatter,
    )



from cg.constants.constants import ControlOptions
from cg.store.models import Case


def are_all_samples_control(case: Case) -> bool:
    """Check if all samples in a case are controls."""

    return all(
        sample.control in [ControlOptions.NEGATIVE, ControlOptions.POSITIVE]
        for sample in case.samples
    )



"""Test for service dtos."""

import pytest

from cg.services.illumina.data_transfer.data_transfer_service import (
    IlluminaSampleSequencingMetricsDTO,
    IlluminaSequencingRunDTO,
)


@pytest.mark.parametrize(
    "percent_value, expected",
    [
        (0.94, 94.0),
        (94.0, 94.0),
        (0.005, 0.5),
    ],
)
def test_percent_validators_sequencing_dto(
    empty_illumina_sequencing_dto: IlluminaSequencingRunDTO, percent_value: float, expected: float
):
    """Test that the percentage validation works."""
    # GIVEN an input value for a DTO percentage attribute

    # WHEN assigning the value to the DTO attributes
    empty_illumina_sequencing_dto.percent_undetermined_reads = percent_value
    empty_illumina_sequencing_dto.percent_q30 = percent_value

    # THEN the value should be converted to a percentage
    assert empty_illumina_sequencing_dto.percent_undetermined_reads == expected
    assert empty_illumina_sequencing_dto.percent_q30 == expected


@pytest.mark.parametrize(
    "percent_value, expected",
    [
        (0.94, 94.0),
        (94.0, 94.0),
        (0.005, 0.5),
    ],
)
def test_percent_validator_sample_sequencing_metrics_dto(
    combined_metric: IlluminaSampleSequencingMetricsDTO,
    percent_value: float,
    expected: float,
):
    """Test that the percentage validation works."""
    # GIVEN an input value for a DTO percentage attribute

    # WHEN assigning the value to the DTO attributes
    combined_metric.base_passing_q30_percent = percent_value

    # THEN the value should be converted to a percentage
    assert combined_metric.base_passing_q30_percent == expected



"""Abstract base classes for order submitters."""

import logging
from abc import ABC, abstractmethod

from cg.models.orders.order import OrderIn
from cg.services.orders.order_lims_service.order_lims_service import OrderLimsService
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class ValidateOrderService(ABC):
    @abstractmethod
    def __init__(self, status_db: Store):
        self.status_db = status_db

    @abstractmethod
    def validate_order(self, order_in: OrderIn):
        pass


class StoreOrderService(ABC):
    @abstractmethod
    def __init__(self, status_db: Store, lims_service: OrderLimsService):
        self.status_db = status_db
        self.lims = lims_service

    @abstractmethod
    def store_order(self, order_in: OrderIn):
        pass

    @staticmethod
    def _fill_in_sample_ids(samples: list[dict], lims_map: dict, id_key: str = "internal_id"):
        """Fill in LIMS sample ids."""
        for sample in samples:
            LOG.debug(f"{sample['name']}: link sample to LIMS")
            if not sample.get(id_key):
                internal_id = lims_map[sample["name"]]
                LOG.info(f"{sample['name']} -> {internal_id}: connect sample to LIMS")
                sample[id_key] = internal_id


class OrderSubmitter(ABC):
    @abstractmethod
    def __init__(
        self,
        validate_order_service: ValidateOrderService,
        store_order_service: StoreOrderService,
    ):
        self.order_validation_service = validate_order_service
        self.order_store_service = store_order_service

    @abstractmethod
    def submit_order(self, order_in: OrderIn) -> dict:
        pass



"""Tests for the deliver base command"""

import logging

from click.testing import CliRunner

from cg.cli.deliver.base import deliver as deliver_cmd
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig


def test_run_base_help(cli_runner: CliRunner):
    """Test to run the deliver base command with --help"""
    # GIVEN a cli runner

    # WHEN running cg deliver help
    result = cli_runner.invoke(deliver_cmd, ["--help"])

    # THEN assert the command exists without problems
    assert result.exit_code == EXIT_SUCCESS
    # THEN assert the information is printed
    assert "Deliver files with CG" in result.output


def test_run_deliver_analysis_help(cli_runner: CliRunner, base_context: CGConfig):
    """Test to run the deliver base command with --help"""
    # GIVEN a cli runner
    # GIVEN a context with store and housekeeper information

    # WHEN running cg deliver help
    result = cli_runner.invoke(deliver_cmd, ["case", "--help"], obj=base_context)

    # THEN assert the command exists without problems
    assert result.exit_code == EXIT_SUCCESS


def test_run_deliver_ticket_help(cli_runner: CliRunner, base_context: CGConfig):
    """Test to run the deliver base command with --help"""
    # GIVEN a cli runner
    # GIVEN a context with store and housekeeper information

    # WHEN running cg deliver help
    result = cli_runner.invoke(deliver_cmd, ["ticket", "--help"], obj=base_context)

    # THEN assert the command exists without problems
    assert result.exit_code == EXIT_SUCCESS


def test_run_deliver_ticket(cli_runner: CliRunner, cg_context: CGConfig, ticket_id):
    """Test for delivering tu customer inbox"""
    # GIVEN a cli runner

    # WHEN running cg deliver ticket
    result = cli_runner.invoke(
        deliver_cmd,
        ["ticket", "--dry-run", "--ticket", ticket_id],
        obj=cg_context,
    )

    # THEN assert that files are delivered
    assert result.exit_code == EXIT_SUCCESS



"""Fixtures for the illumina demux version service."""

from pathlib import Path

import pytest

from cg.services.illumina.file_parsing.demux_version_service import (
    IlluminaDemuxVersionService,
)


@pytest.fixture
def demux_version_file(novaseq_x_demux_runs_dir) -> Path:
    """Return the path to a demultiplexing log file."""
    return Path(
        novaseq_x_demux_runs_dir,
        "dragen-replay.json",
    )


@pytest.fixture
def illumina_demux_version_service() -> IlluminaDemuxVersionService:
    return IlluminaDemuxVersionService()


@pytest.fixture
def expected_demux_software_version() -> str:
    return "4.1.7"


@pytest.fixture
def expected_demux_software() -> str:
    return "dragen"



from cg.services.deliver_files.file_fetcher.models import DeliveryFiles
from cg.services.deliver_files.file_filter.abstract import FilterDeliveryFilesService


class SampleFileFilter(FilterDeliveryFilesService):

    def filter_delivery_files(self, delivery_files: DeliveryFiles, sample_id: str) -> DeliveryFiles:
        delivery_files.sample_files = [
            sample_file
            for sample_file in delivery_files.sample_files
            if sample_file.sample_id == sample_id
        ]
        return delivery_files



from cg.exc import OrderError
from cg.models.orders.order import OrderIn
from cg.models.orders.samples import MetagenomeSample
from cg.services.orders.submitters.order_submitter import ValidateOrderService
from cg.store.models import Customer
from cg.store.store import Store


class ValidateMetagenomeOrderService(ValidateOrderService):

    def __init__(self, status_db: Store):
        self.status_db = status_db

    def validate_order(self, order: OrderIn) -> None:
        self._validate_sample_names_are_unique(samples=order.samples, customer_id=order.customer)

    def _validate_sample_names_are_unique(
        self, samples: list[MetagenomeSample], customer_id: str
    ) -> None:
        """Validate that the names of all samples are unused."""
        customer: Customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=customer_id
        )
        for sample in samples:
            if sample.control:
                continue
            if self.status_db.get_sample_by_customer_and_name(
                customer_entry_id=[customer.id], sample_name=sample.name
            ):
                raise OrderError(f"Sample name {sample.name} already in use")



"""Delivery report specific command options."""

import click

from cg.constants import REPORT_SUPPORTED_WORKFLOW

ARGUMENT_CASE_ID = click.argument(
    "case_id",
    required=False,
    type=str,
)

OPTION_WORKFLOW = click.option(
    "--workflow",
    type=click.Choice(REPORT_SUPPORTED_WORKFLOW),
    help="Limit delivery report generation to a specific workflow",
)

OPTION_STARTED_AT = click.option(
    "--analysis-started-at",
    help="Retrieve analysis started at a specific date (i.e. '2020-05-28  12:00:46')",
)



from cg.exc import OrderError
from cg.models.orders.order import OrderIn
from cg.models.orders.samples import RmlSample
from cg.services.orders.submitters.order_submitter import ValidateOrderService
from cg.store.models import Customer
from cg.store.store import Store


class ValidatePoolOrderService(ValidateOrderService):

    def __init__(self, status_db: Store):
        self.status_db = status_db

    def validate_order(self, order: OrderIn) -> None:
        self._validate_case_names_are_available(
            customer_id=order.customer, samples=order.samples, ticket=order.ticket
        )

    def _validate_case_names_are_available(
        self, customer_id: str, samples: list[RmlSample], ticket: str
    ):
        """Validate names of all samples are not already in use."""
        customer: Customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=customer_id
        )
        for sample in samples:
            case_name: str = self.create_case_name(pool_name=sample.pool, ticket=ticket)
            if self.status_db.get_case_by_name_and_customer(customer=customer, case_name=case_name):
                raise OrderError(
                    f"Case name {case_name} already in use for customer {customer.name}"
                )

    @staticmethod
    def create_case_name(ticket: str, pool_name: str) -> str:
        return f"{ticket}-{pool_name}"



"""Fixtures for cli workflow balsamic tests"""

import datetime as dt
import gzip
from pathlib import Path

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag, Workflow
from cg.constants.constants import CaseActions, FileFormat
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.io.controller import WriteFile
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Sample
from cg.store.store import Store
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.tb_mock import MockTB
from tests.store_helpers import StoreHelpers


@pytest.fixture
def balsamic_dir(tmpdir_factory, apps_dir: Path) -> str:
    """Return the path to the balsamic apps dir"""
    balsamic_dir = tmpdir_factory.mktemp("balsamic")
    return Path(balsamic_dir).absolute().as_posix()


@pytest.fixture
def balsamic_case_id() -> str:
    return "balsamic_case_wgs_single"


@pytest.fixture
def balsamic_housekeeper_dir(tmpdir_factory, balsamic_dir: Path) -> Path:
    """Return the path to the balsamic housekeeper bundle dir."""
    return tmpdir_factory.mktemp("bundles")


@pytest.fixture
def balsamic_pon_1_path(balsamic_dir: Path) -> str:
    balsamic_reference_path = Path(balsamic_dir, "balsamic_bed_1_case_PON_reference.cnn")
    balsamic_reference_path.touch(exist_ok=True)
    return balsamic_reference_path.as_posix()


@pytest.fixture
def balsamic_bed_1_path(balsamic_dir: Path) -> str:
    balsamic_bed_1_path = Path(balsamic_dir, "balsamic_bed_1.bed")
    balsamic_bed_1_path.touch(exist_ok=True)
    return balsamic_bed_1_path.as_posix()


@pytest.fixture
def balsamic_bed_2_path(balsamic_dir: Path) -> str:
    balsamic_bed_2_path = Path(balsamic_dir, "balsamic_bed_2.bed")
    balsamic_bed_2_path.touch(exist_ok=True)
    return balsamic_bed_2_path.as_posix()


@pytest.fixture
def fastq_file_l_1_r_1(balsamic_housekeeper_dir: Path) -> str:
    fastq_filename = Path(
        balsamic_housekeeper_dir, "XXXXXXXXX_000000_S000_L001_R1_001.fastq.gz"
    ).as_posix()
    with gzip.open(fastq_filename, "wb") as wh:
        wh.write(b"@A00689:73:XXXXXXXXX:1:1101:4806:1047 1:N:0:TCCTGGAACA+ACAACCAGTA")
    return fastq_filename


@pytest.fixture
def fastq_file_l_2_r_1(balsamic_housekeeper_dir: Path) -> str:
    fastq_filename = Path(
        balsamic_housekeeper_dir, "XXXXXXXXX_000000_S000_L002_R1_001.fastq.gz"
    ).as_posix()
    with gzip.open(fastq_filename, "wb") as wh:
        wh.write(b"@A00689:73:XXXXXXXXX:2:1101:4806:1047 1:N:0:TCCTGGAACA+ACAACCAGTA")
    return fastq_filename


@pytest.fixture
def fastq_file_l_3_r_1(balsamic_housekeeper_dir: Path) -> str:
    fastq_filename = Path(
        balsamic_housekeeper_dir, "XXXXXXXXX_000000_S000_L003_R1_001.fastq.gz"
    ).as_posix()
    with gzip.open(fastq_filename, "wb") as wh:
        wh.write(b"@A00689:73:XXXXXXXXX:3:1101:4806:1047 1:N:0:TCCTGGAACA+ACAACCAGTA")
    return fastq_filename


@pytest.fixture
def fastq_file_l_4_r_1(balsamic_housekeeper_dir: Path) -> str:
    fastq_filename = Path(
        balsamic_housekeeper_dir, "XXXXXXXXX_000000_S000_L004_R1_001.fastq.gz"
    ).as_posix()
    with gzip.open(fastq_filename, "wb") as wh:
        wh.write(b"@A00689:73:XXXXXXXXX:4:1101:4806:1047 1:N:0:TCCTGGAACA+ACAACCAGTA")
    return fastq_filename


@pytest.fixture
def fastq_file_l_1_r_2(balsamic_housekeeper_dir: Path) -> str:
    fastq_filename = Path(
        balsamic_housekeeper_dir, "XXXXXXXXX_000000_S000_L001_R2_001.fastq.gz"
    ).as_posix()
    with gzip.open(fastq_filename, "wb") as wh:
        wh.write(b"@A00689:73:XXXXXXXXX:1:1101:4806:1047 2:N:0:TCCTGGAACA+ACAACCAGTA")
    return fastq_filename


@pytest.fixture
def fastq_file_l_2_r_2(balsamic_housekeeper_dir: Path) -> str:
    fastq_filename = Path(
        balsamic_housekeeper_dir, "XXXXXXXXX_000000_S000_L002_R2_001.fastq.gz"
    ).as_posix()
    with gzip.open(fastq_filename, "wb") as wh:
        wh.write(b"@A00689:73:XXXXXXXXX:2:1101:4806:1047 2:N:0:TCCTGGAACA+ACAACCAGTA")
    return fastq_filename


@pytest.fixture
def fastq_file_l_3_r_2(balsamic_housekeeper_dir: Path) -> str:
    fastq_filename = Path(
        balsamic_housekeeper_dir, "XXXXXXXXX_000000_S000_L003_R2_001.fastq.gz"
    ).as_posix()
    with gzip.open(fastq_filename, "wb") as wh:
        wh.write(b"@A00689:73:XXXXXXXXX:3:1101:4806:1047 2:N:0:TCCTGGAACA+ACAACCAGTA")
    return fastq_filename


@pytest.fixture
def fastq_file_l_4_r_2(balsamic_housekeeper_dir: Path) -> str:
    fastq_filename = Path(
        balsamic_housekeeper_dir, "XXXXXXXXX_000000_S000_L004_R2_001.fastq.gz"
    ).as_posix()
    with gzip.open(fastq_filename, "wb") as wh:
        wh.write(b"@A00689:73:XXXXXXXXX:4:1101:4806:1047 2:N:0:TCCTGGAACA+ACAACCAGTA")
    return fastq_filename


@pytest.fixture
def balsamic_mock_fastq_files(
    fastq_file_l_1_r_1: Path,
    fastq_file_l_1_r_2: Path,
    fastq_file_l_2_r_1: Path,
    fastq_file_l_2_r_2: Path,
    fastq_file_l_3_r_1: Path,
    fastq_file_l_3_r_2: Path,
    fastq_file_l_4_r_1: Path,
    fastq_file_l_4_r_2: Path,
) -> list:
    """Return list of all mock fastq files to commmit to mock housekeeper"""
    return [
        fastq_file_l_1_r_1,
        fastq_file_l_1_r_2,
        fastq_file_l_2_r_1,
        fastq_file_l_2_r_2,
        fastq_file_l_3_r_1,
        fastq_file_l_3_r_2,
        fastq_file_l_4_r_1,
        fastq_file_l_4_r_2,
    ]


@pytest.fixture(scope="function")
def balsamic_housekeeper(housekeeper_api, helpers, balsamic_mock_fastq_files: list):
    """Create populated housekeeper that holds files for all mock samples"""

    samples = [
        "sample_case_wgs_paired_tumor",
        "sample_case_wgs_paired_normal",
        "sample_case_tgs_paired_tumor",
        "sample_case_tgs_paired_normal",
        "sample_case_wgs_single_tumor",
        "sample_case_tgs_single_tumor",
        "sample_case_tgs_single_normal_error",
        "sample_case_tgs_paired_tumor_error",
        "sample_case_tgs_paired_tumor2_error",
        "sample_case_tgs_paired_normal_error",
        "mixed_sample_case_wgs_paired_tumor_error",
        "mixed_sample_case_tgs_paired_normal_error",
        "mixed_sample_case_mixed_bed_paired_tumor_error",
        "mixed_sample_case_mixed_bed_paired_normal_error",
        "mip_sample_case_wgs_single_tumor",
        "sample_case_wgs_paired_two_normal_tumor_error",
        "sample_case_wgs_paired_two_normal_normal1_error",
        "sample_case_wgs_paired_two_normal_normal2_error",
        "sample_case_wes_panel_error",
        "sample_case_wes_tumor",
    ]

    for sample in samples:
        bundle_data = {
            "name": sample,
            "created": dt.datetime.now(),
            "version": "1.0",
            "files": [
                {"path": f, "tags": [SequencingFileTag.FASTQ], "archive": False}
                for f in balsamic_mock_fastq_files
            ],
        }
        helpers.ensure_hk_bundle(store=housekeeper_api, bundle_data=bundle_data)
    return housekeeper_api


@pytest.fixture
def balsamic_lims(context_config: dict) -> MockLimsAPI:
    """Create populated mock LIMS api to mimic all functionality of LIMS used by BALSAMIC"""

    balsamic_lims = MockLimsAPI(context_config)
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_wgs_paired_tumor",
        capture_kit=None,
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_wgs_paired_normal",
        capture_kit=None,
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_tgs_paired_tumor",
        capture_kit="BalsamicBed1",
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_tgs_paired_normal",
        capture_kit="BalsamicBed1",
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_wgs_single_tumor",
        capture_kit=None,
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_tgs_single_tumor",
        capture_kit="BalsamicBed1",
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_tgs_single_normal_error",
        capture_kit="BalsamicBed1",
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_tgs_paired_tumor_error",
        capture_kit="BalsamicBed1",
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_tgs_paired_tumor2_error",
        capture_kit="BalsamicBed1",
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_tgs_paired_normal_error",
        capture_kit="BalsamicBed1",
    )
    balsamic_lims.add_capture_kit(
        internal_id="mixed_sample_case_wgs_paired_tumor_error",
        capture_kit=None,
    )
    balsamic_lims.add_capture_kit(
        internal_id="mixed_sample_case_tgs_paired_normal_error",
        capture_kit="BalsamicBed1",
    )
    balsamic_lims.add_capture_kit(
        internal_id="mixed_sample_case_mixed_bed_paired_tumor_error",
        capture_kit="BalsamicBed1",
    )
    balsamic_lims.add_capture_kit(
        internal_id="mixed_sample_case_mixed_bed_paired_normal_error",
        capture_kit="BalsamicBed2",
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_wes_tumor",
        capture_kit="BalsamicBed2",
    )

    balsamic_lims.add_capture_kit(
        internal_id="mip_sample_case_wgs_single_tumor",
        capture_kit=None,
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_wgs_paired_two_normal_tumor_error",
        capture_kit=None,
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_wgs_paired_two_normal_normal1_error",
        capture_kit=None,
    )
    balsamic_lims.add_capture_kit(
        internal_id="sample_case_wgs_paired_two_normal_normal2_error",
        capture_kit=None,
    )

    return balsamic_lims


@pytest.fixture(scope="function")
def balsamic_context(
    cg_context: CGConfig,
    helpers: StoreHelpers,
    balsamic_lims: MockLimsAPI,
    balsamic_housekeeper: HousekeeperAPI,
    trailblazer_api: MockTB,
    cg_dir,
) -> CGConfig:
    """context to use in cli"""
    cg_context.housekeeper_api_ = balsamic_housekeeper
    cg_context.lims_api_ = balsamic_lims
    cg_context.trailblazer_api_ = trailblazer_api
    cg_context.meta_apis["analysis_api"] = BalsamicAnalysisAPI(config=cg_context)
    status_db: Store = cg_context.status_db

    # Create tgs application version
    helpers.ensure_application_version(
        store=status_db,
        application_tag="TGSA",
        prep_category=SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
        target_reads=10,
    )

    # Create wes application version
    helpers.ensure_application_version(
        store=status_db,
        application_tag="WESA",
        prep_category=SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING,
        target_reads=10,
    )

    # Create wgs application version
    helpers.ensure_application_version(
        store=status_db,
        application_tag="WGSA",
        prep_category=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        target_reads=10,
    )

    # Create a mic application version
    helpers.ensure_application_version(
        store=status_db,
        application_tag="MICA",
        prep_category="mic",
        target_reads=10,
    )

    # Create textbook case for WHOLE_GENOME_SEQUENCING PAIRED with enough reads
    case_wgs_paired_enough_reads = helpers.add_case(
        store=status_db,
        internal_id="balsamic_case_wgs_paired_enough_reads",
        name="balsamic_case_wgs_paired_enough_reads",
        data_analysis=Workflow.BALSAMIC,
        action=CaseActions.HOLD,
    )
    sample_case_wgs_paired_tumor_enough_reads = helpers.add_sample(
        status_db,
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        is_tumour=True,
        internal_id="sample_case_wgs_paired_tumor_enough_reads",
        reads=10,
        last_sequenced_at=dt.datetime.now(),
    )
    sample_case_wgs_paired_normal_enough_reads = helpers.add_sample(
        status_db,
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        is_tumour=False,
        internal_id="sample_case_wgs_paired_normal_enough_reads",
        reads=10,
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(
        status_db,
        case=case_wgs_paired_enough_reads,
        sample=sample_case_wgs_paired_tumor_enough_reads,
    )
    helpers.add_relationship(
        status_db,
        case=case_wgs_paired_enough_reads,
        sample=sample_case_wgs_paired_normal_enough_reads,
    )

    # Create textbook case for WHOLE_GENOME_SEQUENCING PAIRED
    case_wgs_paired = helpers.add_case(
        store=status_db,
        internal_id="balsamic_case_wgs_paired",
        name="balsamic_case_wgs_paired",
        data_analysis=Workflow.BALSAMIC,
        action=CaseActions.HOLD,
    )
    sample_case_wgs_paired_tumor = helpers.add_sample(
        status_db,
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        is_tumour=True,
        internal_id="sample_case_wgs_paired_tumor",
        reads=10,
        last_sequenced_at=dt.datetime.now(),
    )
    sample_case_wgs_paired_normal = helpers.add_sample(
        status_db,
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        is_tumour=False,
        internal_id="sample_case_wgs_paired_normal",
        reads=10,
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(status_db, case=case_wgs_paired, sample=sample_case_wgs_paired_tumor)
    helpers.add_relationship(status_db, case=case_wgs_paired, sample=sample_case_wgs_paired_normal)

    # Create textbook case for TARGETED_GENOME_SEQUENCING PAIRED without enough reads
    case_tgs_paired = helpers.add_case(
        status_db,
        internal_id="balsamic_case_tgs_paired",
        name="balsamic_case_tgs_paired",
        data_analysis=Workflow.BALSAMIC,
    )
    sample_case_tgs_paired_tumor = helpers.add_sample(
        status_db,
        application_tag="TGSA",
        application_type=SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
        is_tumour=True,
        internal_id="sample_case_tgs_paired_tumor",
        reads=10,
        last_sequenced_at=dt.datetime.now(),
    )
    sample_case_tgs_paired_normal = helpers.add_sample(
        status_db,
        application_tag="TGSA",
        application_type=SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
        is_tumour=False,
        internal_id="sample_case_tgs_paired_normal",
        reads=0,
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(status_db, case=case_tgs_paired, sample=sample_case_tgs_paired_tumor)
    helpers.add_relationship(status_db, case=case_tgs_paired, sample=sample_case_tgs_paired_normal)

    # Create textbook case for WHOLE_GENOME_SEQUENCING TUMOR ONLY
    case_wgs_single = helpers.add_case(
        status_db,
        internal_id="balsamic_case_wgs_single",
        name="balsamic_case_wgs_single",
        data_analysis=Workflow.BALSAMIC,
    )
    sample_case_wgs_single_tumor = helpers.add_sample(
        status_db,
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        is_tumour=True,
        internal_id="sample_case_wgs_single_tumor",
        reads=100,
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(status_db, case=case_wgs_single, sample=sample_case_wgs_single_tumor)

    # Create textbook case for TARGETED_GENOME_SEQUENCING TUMOR ONLY
    case_tgs_single = helpers.add_case(
        status_db,
        internal_id="balsamic_case_tgs_single",
        name="balsamic_case_tgs_single",
        data_analysis=Workflow.BALSAMIC,
    )
    sample_case_tgs_single_tumor = helpers.add_sample(
        status_db,
        application_tag="TGSA",
        application_type=SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
        is_tumour=True,
        internal_id="sample_case_tgs_single_tumor",
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(status_db, case=case_tgs_single, sample=sample_case_tgs_single_tumor)

    # Create ERROR case for TARGETED_GENOME_SEQUENCING NORMAL ONLY
    case_tgs_single_error = helpers.add_case(
        status_db,
        internal_id="balsamic_case_tgs_single_error",
        name="balsamic_case_tgs_single_error",
        data_analysis=Workflow.BALSAMIC,
    )
    sample_case_tgs_single_normal_error = helpers.add_sample(
        status_db,
        application_tag="TGSA",
        application_type=SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
        is_tumour=False,
        internal_id="sample_case_tgs_single_normal_error",
        reads=0,
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(
        status_db,
        case=case_tgs_single_error,
        sample=sample_case_tgs_single_normal_error,
    )

    # Create ERROR case for TARGETED_GENOME_SEQUENCING TWO TUMOR ONE NORMAL
    case_tgs_paired_error = helpers.add_case(
        status_db,
        internal_id="balsamic_case_tgs_paired_error",
        name="balsamic_case_tgs_paired_error",
        data_analysis=Workflow.BALSAMIC,
    )
    sample_case_tgs_paired_tumor_error = helpers.add_sample(
        status_db,
        application_tag="TGSA",
        application_type=SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
        is_tumour=True,
        reads=0,
        internal_id="sample_case_tgs_paired_tumor_error",
        last_sequenced_at=dt.datetime.now(),
    )
    sample_case_tgs_paired_tumor2_error = helpers.add_sample(
        status_db,
        application_tag="TGSA",
        application_type=SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
        is_tumour=True,
        reads=0,
        internal_id="sample_case_tgs_paired_tumor2_error",
        last_sequenced_at=dt.datetime.now(),
    )
    sample_case_tgs_paired_normal_error = helpers.add_sample(
        status_db,
        application_tag="TGSA",
        application_type=SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
        is_tumour=False,
        reads=0,
        internal_id="sample_case_tgs_paired_normal_error",
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(
        status_db,
        case=case_tgs_paired_error,
        sample=sample_case_tgs_paired_tumor_error,
    )
    helpers.add_relationship(
        status_db,
        case=case_tgs_paired_error,
        sample=sample_case_tgs_paired_tumor2_error,
    )
    helpers.add_relationship(
        status_db,
        case=case_tgs_paired_error,
        sample=sample_case_tgs_paired_normal_error,
    )

    # Create ERROR case for MIXED application type
    case_mixed_paired_error = helpers.add_case(
        status_db,
        internal_id="balsamic_case_mixed_paired_error",
        name="balsamic_case_mixed_paired_error",
        data_analysis=Workflow.BALSAMIC,
    )
    mixed_sample_case_wgs_paired_tumor_error = helpers.add_sample(
        status_db,
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        is_tumour=True,
        reads=0,
        internal_id="mixed_sample_case_wgs_paired_tumor_error",
        last_sequenced_at=dt.datetime.now(),
    )
    mixed_sample_case_tgs_paired_normal_error = helpers.add_sample(
        status_db,
        application_tag="TGSA",
        application_type=SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
        is_tumour=False,
        reads=0,
        internal_id="mixed_sample_case_tgs_paired_normal_error",
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(
        status_db,
        case=case_mixed_paired_error,
        sample=mixed_sample_case_wgs_paired_tumor_error,
    )
    helpers.add_relationship(
        status_db,
        case=case_mixed_paired_error,
        sample=mixed_sample_case_tgs_paired_normal_error,
    )

    # Create ERROR case for MIXED application type NOT BALSAMIC APPLICATION
    case_mixed_wgs_mic_paired_error = helpers.add_case(
        status_db,
        internal_id="balsamic_case_mixed_wgs_mic_paired_error",
        name="balsamic_case_mixed_wgs_mic_paired_error",
        data_analysis=Workflow.BALSAMIC,
    )
    mixed_sample_case_wgs_mic_paired_tumor_error = helpers.add_sample(
        status_db,
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        is_tumour=True,
        internal_id="mixed_sample_case_wgs_mic_paired_tumor_error",
        reads=0,
        last_sequenced_at=dt.datetime.now(),
    )
    mixed_sample_case_wgs_mic_paired_normal_error: Sample = helpers.add_sample(
        status_db,
        application_tag="MICA",
        application_type="mic",
        is_tumour=False,
        internal_id="mixed_sample_case_wgs_mic_paired_normal_error",
        reads=0,
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(
        status_db,
        case=case_mixed_wgs_mic_paired_error,
        sample=mixed_sample_case_wgs_mic_paired_tumor_error,
    )
    helpers.add_relationship(
        status_db,
        case=case_mixed_wgs_mic_paired_error,
        sample=mixed_sample_case_wgs_mic_paired_normal_error,
    )

    # Create ERROR case for MIXED TARGET BED
    case_mixed_bed_paired_error = helpers.add_case(
        status_db,
        internal_id="balsamic_case_mixed_bed_paired_error",
        name="balsamic_case_mixed_bed_paired_error",
        data_analysis=Workflow.BALSAMIC,
    )
    mixed_sample_case_mixed_bed_paired_tumor_error = helpers.add_sample(
        status_db,
        application_tag="TGSA",
        application_type=SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
        is_tumour=True,
        internal_id="mixed_sample_case_mixed_bed_paired_tumor_error",
        last_sequenced_at=dt.datetime.now(),
    )
    mixed_sample_case_mixed_bed_paired_normal_error = helpers.add_sample(
        status_db,
        application_tag="TGSA",
        application_type=SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
        is_tumour=False,
        internal_id="mixed_sample_case_mixed_bed_paired_normal_error",
        reads=0,
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(
        status_db,
        case=case_mixed_bed_paired_error,
        sample=mixed_sample_case_mixed_bed_paired_tumor_error,
    )
    helpers.add_relationship(
        status_db,
        case=case_mixed_bed_paired_error,
        sample=mixed_sample_case_mixed_bed_paired_normal_error,
    )

    # Create ERROR case for WHOLE_GENOME_SEQUENCING TUMOR ONLY MIP CLI_OPTION_ANALYSIS ONLY
    mip_case_wgs_single = helpers.add_case(
        status_db,
        internal_id="mip_case_wgs_single",
        name="mip_case_wgs_single",
        data_analysis=Workflow.MIP_DNA,
    )
    mip_sample_case_wgs_single_tumor = helpers.add_sample(
        status_db,
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        is_tumour=True,
        internal_id="mip_sample_case_wgs_single_tumor",
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(
        status_db,
        case=mip_case_wgs_single,
        sample=mip_sample_case_wgs_single_tumor,
    )

    # Create ERROR case for WHOLE_GENOME_SEQUENCING ONE TUMOR TWO NORMAL
    case_wgs_paired_two_normal_error = helpers.add_case(
        status_db,
        internal_id="balsamic_case_wgs_paired_two_normal_error",
        name="balsamic_case_wgs_paired_two_normal_error",
        data_analysis=Workflow.BALSAMIC,
    )
    sample_case_wgs_paired_two_normal_tumor_error = helpers.add_sample(
        status_db,
        application_tag="WGSA",
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        is_tumour=True,
        internal_id="sample_case_wgs_paired_two_normal_tumor_error",
        reads=0,
        last_sequenced_at=dt.datetime.now(),
    )
    sample_case_wgs_paired_two_normal_normal1_error = helpers.add_sample(
        status_db,
        application_tag="WGSA",
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        is_tumour=False,
        internal_id="sample_case_wgs_paired_two_normal_normal1_error",
        reads=0,
        last_sequenced_at=dt.datetime.now(),
    )
    sample_case_wgs_paired_two_normal_normal2_error = helpers.add_sample(
        status_db,
        application_tag="WGSA",
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        is_tumour=False,
        internal_id="sample_case_wgs_paired_two_normal_normal2_error",
        reads=0,
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(
        status_db,
        case=case_wgs_paired_two_normal_error,
        sample=sample_case_wgs_paired_two_normal_tumor_error,
    )
    helpers.add_relationship(
        status_db,
        case=case_wgs_paired_two_normal_error,
        sample=sample_case_wgs_paired_two_normal_normal1_error,
    )
    helpers.add_relationship(
        status_db,
        case=case_wgs_paired_two_normal_error,
        sample=sample_case_wgs_paired_two_normal_normal2_error,
    )

    # Create WHOLE_EXOME_SEQUENCING case with 1 tumor sample
    case_wes_tumor = helpers.add_case(
        status_db,
        internal_id="balsamic_case_wes_tumor",
        name="balsamic_case_wes_tumor",
        data_analysis=Workflow.BALSAMIC,
    )
    sample_case_wes_tumor = helpers.add_sample(
        status_db,
        application_tag="WESA",
        application_type=SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING,
        is_tumour=True,
        internal_id="sample_case_wes_tumor",
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(status_db, case=case_wes_tumor, sample=sample_case_wes_tumor)

    # Create ERROR case for WHOLE_EXOME_SEQUENCING when no panel is found
    case_wes_panel_error = helpers.add_case(
        status_db,
        internal_id="balsamic_case_wes_panel_error",
        name="balsamic_case_wes_panel_error",
        data_analysis=Workflow.BALSAMIC,
    )
    sample_case_wes_panel_error = helpers.add_sample(
        status_db,
        application_tag="WESA",
        application_type=SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING,
        is_tumour=True,
        internal_id="sample_case_wes_panel_error",
        reads=0,
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_relationship(
        status_db, case=case_wes_panel_error, sample=sample_case_wes_panel_error
    )

    # Create ERROR case with NO SAMPLES
    helpers.add_case(status_db, internal_id="no_sample_case", name="no_sample_case")

    # Create BED1 version 1
    bed1_name = "BalsamicBed1"
    bed1_filename = "balsamic_bed_1.bed"
    Path(cg_dir, bed1_filename).touch(exist_ok=True)
    bed1 = status_db.add_bed(name=bed1_name)
    status_db.session.add(bed1)
    version1 = status_db.add_bed_version(
        bed=bed1, version=1, filename=bed1_filename, shortname=bed1_name
    )
    status_db.session.add(version1)

    # Create BED2 version 1
    bed2_name = "BalsamicBed2"
    bed2_filename = "balsamic_bed_2.bed"
    Path(cg_dir, bed2_filename).touch(exist_ok=True)
    bed2 = status_db.add_bed(name=bed2_name)
    status_db.session.add(bed2)
    version2 = status_db.add_bed_version(
        bed=bed2, version=1, filename=bed2_filename, shortname=bed2_name
    )
    status_db.session.add(version2)
    status_db.session.commit()
    return cg_context


@pytest.fixture
def mock_config(balsamic_dir: Path, balsamic_case_id: str) -> None:
    """Create dummy config file at specified path"""

    config_data = {
        "analysis": {
            "case_id": f"{balsamic_case_id}",
            "analysis_type": "paired",
            "sequencing_type": "targeted",
            "analysis_dir": f"{balsamic_dir}",
            "fastq_path": f"{balsamic_dir}/{balsamic_case_id}/analysis/fastq/",
            "script": f"{balsamic_dir}/{balsamic_case_id}/scripts/",
            "log": f"{balsamic_dir}/{balsamic_case_id}/logs/",
            "result": f"{balsamic_dir}/{balsamic_case_id}/analysis",
            "benchmark": f"{balsamic_dir}/{balsamic_case_id}/benchmarks/",
            "dag": f"{balsamic_dir}/{balsamic_case_id}/{balsamic_case_id}_BALSAMIC_4.4.0_graph.pdf",
            "BALSAMIC_version": "4",
            "config_creation_date": "2020-07-15 17:35",
        }
    }
    Path.mkdir(Path(balsamic_dir, balsamic_case_id), parents=True, exist_ok=True)
    WriteFile.write_file_from_content(
        content=config_data,
        file_format=FileFormat.JSON,
        file_path=Path(balsamic_dir, balsamic_case_id, balsamic_case_id + ".json"),
    )


@pytest.fixture
def deliverable_data(balsamic_dir: Path, balsamic_case_id: str) -> dict:
    samples = [
        "sample_case_wgs_single_tumor",
    ]

    return {
        "files": [
            {
                "path": f"{balsamic_dir}/{balsamic_case_id}/multiqc_report.html",
                "path_index": "",
                "step": "multiqc",
                "tag": ["qc"],
                "id": "T_WGS",
                "format": "html",
                "mandatory": True,
            },
            {
                "path": f"{balsamic_dir}/{balsamic_case_id}/concatenated_{samples[0]}_R_1.fp.fastq.gz",
                "path_index": "",
                "step": "fastp",
                "tag": [f"concatenated_{samples[0]}_R", "qc"],
                "id": f"concatenated_{samples[0]}_R",
                "format": "fastq.gz",
                "mandatory": True,
            },
            {
                "path": f"{balsamic_dir}/{balsamic_case_id}/CNV.somatic.{balsamic_case_id}.cnvkit.pass.vcf.gz.tbi",
                "path_index": "",
                "step": "vep_somatic",
                "format": "vcf.gz.tbi",
                "tag": [
                    "CNV",
                    balsamic_case_id,
                    "cnvkit",
                    "annotation",
                    "somatic",
                    "index",
                ],
                "id": balsamic_case_id,
                "mandatory": True,
            },
        ]
    }


@pytest.fixture
def mock_deliverable(balsamic_dir: Path, deliverable_data: dict, balsamic_case_id: str) -> None:
    """Create deliverable file with dummy data and files to deliver"""
    Path.mkdir(
        Path(balsamic_dir, balsamic_case_id, "analysis", "delivery_report"),
        parents=True,
        exist_ok=True,
    )
    for report_entry in deliverable_data["files"]:
        Path(report_entry["path"]).touch(exist_ok=True)
    WriteFile.write_file_from_content(
        content=deliverable_data,
        file_format=FileFormat.JSON,
        file_path=Path(
            balsamic_dir, balsamic_case_id, "analysis", "delivery_report", balsamic_case_id + ".hk"
        ),
    )


@pytest.fixture
def hermes_deliverables(deliverable_data: dict, balsamic_case_id: str) -> dict:
    hermes_output: dict = {"workflow": "balsamic", "bundle_id": balsamic_case_id, "files": []}
    for file_info in deliverable_data["files"]:
        tags: list[str] = []
        if "html" in file_info["format"]:
            tags.append("multiqc-html")
        elif "fastq" in file_info["format"]:
            tags.append("fastq")
        elif "vcf" in file_info["format"]:
            tags.extend(["vcf-snv-clinical", "cnvkit", "filtered"])
        hermes_output["files"].append({"path": file_info["path"], "tags": tags, "mandatory": True})
    return hermes_output


@pytest.fixture
def malformed_hermes_deliverables(hermes_deliverables: dict) -> dict:
    malformed_deliverable = hermes_deliverables.copy()
    malformed_deliverable.pop("workflow")

    return malformed_deliverable


@pytest.fixture
def mock_analysis_finish(balsamic_dir: Path, balsamic_case_id: str) -> None:
    """Create analysis_finish file for testing"""
    Path.mkdir(Path(balsamic_dir, balsamic_case_id, "analysis"), parents=True, exist_ok=True)
    Path(balsamic_dir, balsamic_case_id, "analysis", "analysis_finish").touch(exist_ok=True)



"""Chanjo2 client API utility methods."""

import logging
from functools import wraps

from pydantic import ValidationError
from requests import RequestException

from cg.exc import Chanjo2APIClientError, Chanjo2RequestError, Chanjo2ResponseError

LOG = logging.getLogger(__name__)


def handle_client_errors(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RequestException as error:
            LOG.error(f"Request against Chanjo2 API failed: {error}")
            raise Chanjo2RequestError(error) from error
        except ValidationError as error:
            LOG.error(f"Invalidly formatted response from Chanjo2 API: {error}")
            raise Chanjo2ResponseError(error) from error
        except Exception as error:
            LOG.error(f"Unexpected error in Chanjo2 API: {error}")
            raise Chanjo2APIClientError(error) from error

    return wrapper



"""NF Analysis upload API."""

import datetime as dt
import logging

import click

from cg.cli.generate.delivery_report.base import generate_delivery_report
from cg.cli.upload.scout import upload_to_scout
from cg.constants import (
    REPORT_SUPPORTED_DATA_DELIVERY,
    REPORT_SUPPORTED_WORKFLOW,
    DataDelivery,
    Workflow,
)
from cg.meta.upload.upload_api import UploadAPI
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Analysis, Case

LOG = logging.getLogger(__name__)


class NfAnalysisUploadAPI(UploadAPI):
    """Nf analysis upload API."""

    def __init__(self, config: CGConfig, workflow: Workflow):
        self.analysis_api: NfAnalysisAPI = NfAnalysisAPI(config, workflow)
        super().__init__(config=config, analysis_api=self.analysis_api)

    def upload(self, ctx: click.Context, case: Case, restart: bool) -> None:
        """Upload NF analysis data and files."""
        analysis: Analysis = case.analyses[0]
        self.update_upload_started_at(analysis=analysis)

        # Delivery report generation
        if (
            case.data_analysis in REPORT_SUPPORTED_WORKFLOW
            and case.data_delivery in REPORT_SUPPORTED_DATA_DELIVERY
        ):
            ctx.invoke(generate_delivery_report, case_id=case.internal_id)

        # Clinical delivery
        self.upload_files_to_customer_inbox(case)

        # Scout specific upload
        if DataDelivery.SCOUT in case.data_delivery:
            ctx.invoke(upload_to_scout, case_id=case.internal_id, re_upload=restart)
        LOG.info(
            f"Upload of case {case.internal_id} was successful. Setting uploaded at to {dt.datetime.now()}"
        )
        self.update_uploaded_at(analysis=analysis)



import logging
from pathlib import Path

from cg.apps.tb import TrailblazerAPI
from cg.apps.tb.models import TrailblazerAnalysis
from cg.constants import Priority, Workflow
from cg.constants.tb import AnalysisTypes
from cg.services.analysis_service.analysis_service import AnalysisService
from cg.services.deliver_files.deliver_files_service.error_handling import (
    handle_no_delivery_files_error,
)
from cg.services.deliver_files.file_fetcher.abstract import FetchDeliveryFilesService
from cg.services.deliver_files.file_fetcher.models import DeliveryFiles
from cg.services.deliver_files.file_filter.abstract import FilterDeliveryFilesService
from cg.services.deliver_files.file_formatter.abstract import DeliveryFileFormattingService
from cg.services.deliver_files.file_formatter.models import FormattedFiles
from cg.services.deliver_files.file_mover.service import DeliveryFilesMover
from cg.services.deliver_files.rsync.service import DeliveryRsyncService
from cg.store.exc import EntryNotFoundError
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class DeliverFilesService:
    """
    Deliver files to the customer inbox on the HPC and Rsync them to the inbox folder on the delivery server.
    1. Get the files to deliver from Housekeeper based on workflow and data delivery
    2. Create a delivery folder structure in the customer folder on Hasta and move the files there
    3. Reformatting of output / renaming of files
    4. Rsync the files to the customer inbox on the delivery server
    5. Add the rsync job to Trailblazer
    """

    def __init__(
        self,
        delivery_file_manager_service: FetchDeliveryFilesService,
        file_filter: FilterDeliveryFilesService,
        move_file_service: DeliveryFilesMover,
        file_formatter_service: DeliveryFileFormattingService,
        rsync_service: DeliveryRsyncService,
        tb_service: TrailblazerAPI,
        analysis_service: AnalysisService,
        status_db: Store,
    ):
        self.file_manager = delivery_file_manager_service
        self.file_filter = file_filter
        self.file_mover = move_file_service
        self.file_formatter = file_formatter_service
        self.status_db = status_db
        self.rsync_service = rsync_service
        self.tb_service = tb_service
        self.analysis_service = analysis_service

    @handle_no_delivery_files_error
    def deliver_files_for_case(
        self, case: Case, delivery_base_path: Path, dry_run: bool = False
    ) -> None:
        """Deliver the files for a case to the customer folder."""
        delivery_files: DeliveryFiles = self.file_manager.get_files_to_deliver(
            case_id=case.internal_id
        )
        moved_files: DeliveryFiles = self.file_mover.move_files(
            delivery_files=delivery_files, delivery_base_path=delivery_base_path
        )
        formatted_files: FormattedFiles = self.file_formatter.format_files(moved_files)
        folders_to_deliver: set[Path] = set(
            [formatted_file.formatted_path.parent for formatted_file in formatted_files.files]
        )
        job_id: int = self._start_rsync_job(
            case=case, dry_run=dry_run, folders_to_deliver=folders_to_deliver
        )
        self._add_trailblazer_tracking(case=case, job_id=job_id, dry_run=dry_run)

    def deliver_files_for_ticket(
        self, ticket_id: str, delivery_base_path: Path, dry_run: bool = False
    ) -> None:
        """Deliver the files for all cases in a ticket to the customer folder."""
        cases: list[Case] = self.status_db.get_cases_by_ticket_id(ticket_id)
        if not cases:
            raise EntryNotFoundError(f"No cases found for ticket {ticket_id}")
        for case in cases:
            self.deliver_files_for_case(
                case=case, delivery_base_path=delivery_base_path, dry_run=dry_run
            )

    def deliver_files_for_sample(
        self, case: Case, sample_id: str, delivery_base_path: Path, dry_run: bool = False
    ):
        """Deliver the files for a sample to the customer folder."""
        delivery_files: DeliveryFiles = self.file_manager.get_files_to_deliver(
            case_id=case.internal_id
        )
        filtered_files: DeliveryFiles = self.file_filter.filter_delivery_files(
            delivery_files=delivery_files, sample_id=sample_id
        )
        moved_files: DeliveryFiles = self.file_mover.move_files(
            delivery_files=filtered_files, delivery_base_path=delivery_base_path
        )
        formatted_files: FormattedFiles = self.file_formatter.format_files(moved_files)
        folders_to_deliver: set[Path] = set(
            [formatted_file.formatted_path.parent for formatted_file in formatted_files.files]
        )
        job_id: int = self._start_rsync_job(
            case=case, dry_run=dry_run, folders_to_deliver=folders_to_deliver
        )
        self._add_trailblazer_tracking(case=case, job_id=job_id, dry_run=dry_run)

    def _start_rsync_job(self, case: Case, dry_run: bool, folders_to_deliver: set[Path]) -> int:
        LOG.debug(f"[RSYNC] Starting rsync job for case {case.internal_id}")
        job_id: int = self.rsync_service.run_rsync_for_case(
            case=case,
            dry_run=dry_run,
            folders_to_deliver=folders_to_deliver,
        )
        self.rsync_service.write_trailblazer_config(
            content={"jobs": [str(job_id)]},
            config_path=self.rsync_service.trailblazer_config_path,
            dry_run=dry_run,
        )
        return job_id

    def _add_trailblazer_tracking(self, case: Case, job_id: int, dry_run: bool) -> None:
        if dry_run:
            LOG.info(f"Would have added the analysis for case {case.internal_id} to Trailblazer")
        else:
            LOG.debug(f"[TB SERVICE] Adding analysis for case {case.internal_id} to Trailblazer")
            analysis: TrailblazerAnalysis = self.tb_service.add_pending_analysis(
                case_id=f"{case.internal_id}_rsync",
                analysis_type=AnalysisTypes.OTHER,
                config_path=self.rsync_service.trailblazer_config_path.as_posix(),
                order_id=case.latest_order.id,
                out_dir=self.rsync_service.log_dir.as_posix(),
                slurm_quality_of_service=Priority.priority_to_slurm_qos().get(case.priority),
                workflow=Workflow.RSYNC,
                ticket=case.latest_ticket,
            )
            self.tb_service.add_upload_job_to_analysis(analysis_id=analysis.id, slurm_id=job_id)
            self.analysis_service.add_upload_job(slurm_id=job_id, case_id=case.internal_id)
            LOG.info(f"Transfer of case {case.internal_id} started with SLURM job id {job_id}")



import os
import pytest
from pathlib import Path

from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)
from cg.services.deliver_files.file_fetcher.models import (
    CaseFile,
    SampleFile,
)
from cg.services.deliver_files.file_formatter.models import FormattedFile
from cg.services.deliver_files.file_formatter.utils.case_service import (
    CaseFileFormatter,
)
from cg.services.deliver_files.file_formatter.utils.sample_concatenation_service import (
    SampleFileConcatenationFormatter,
)
from cg.services.deliver_files.file_formatter.utils.sample_service import (
    SampleFileFormatter,
    FileManagingService,
    SampleFileNameFormatter,
)


@pytest.mark.parametrize(
    "moved_files,expected_formatted_files,file_formatter",
    [
        (
            "expected_moved_analysis_case_delivery_files",
            "expected_formatted_analysis_case_files",
            CaseFileFormatter(),
        ),
        (
            "expected_moved_analysis_sample_delivery_files",
            "expected_formatted_analysis_sample_files",
            SampleFileFormatter(
                file_manager=FileManagingService(), file_name_formatter=SampleFileNameFormatter()
            ),
        ),
        (
            "fastq_concatenation_sample_files",
            "expected_concatenated_fastq_formatted_files",
            SampleFileConcatenationFormatter(
                file_manager=FileManagingService(),
                file_formatter=SampleFileNameFormatter(),
                concatenation_service=FastqConcatenationService(),
            ),
        ),
    ],
)
def test_file_formatter_utils(
    moved_files: list[CaseFile | SampleFile],
    expected_formatted_files: list[FormattedFile],
    file_formatter: CaseFileFormatter | SampleFileFormatter | SampleFileConcatenationFormatter,
    request,
):
    # GIVEN existing case files, a case file formatter and a ticket directory path and a customer inbox
    moved_files: list[CaseFile | SampleFile] = request.getfixturevalue(moved_files)
    expected_formatted_files: list[FormattedFile] = request.getfixturevalue(
        expected_formatted_files
    )
    ticket_dir_path: Path = moved_files[0].file_path.parent

    os.makedirs(ticket_dir_path, exist_ok=True)

    for moved_file in moved_files:
        moved_file.file_path.touch()

    # WHEN formatting the case files
    formatted_files: list[FormattedFile] = file_formatter.format_files(
        moved_files=moved_files,
        ticket_dir_path=ticket_dir_path,
    )

    # THEN the case files should be formatted
    assert formatted_files == expected_formatted_files
    for file in formatted_files:
        assert file.formatted_path.exists()
        assert not file.original_path.exists()



"""Test CLI function to get sequencing runs from StatusDB."""

from datetime import datetime

from click.testing import CliRunner

from cg.cli.get import get
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.models.cg_config import CGConfig
from cg.store.models import IlluminaFlowCell, IlluminaSequencingRun
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_get_sequencing_run_wrong_device_id(cli_runner: CliRunner, base_context: CGConfig):
    """Test that trying to get a sequencing run using a non-existing device id fails."""
    # GIVEN a database

    # WHEN getting a non-existent sequencing run
    result = cli_runner.invoke(get, ["sequencing-run", "does_not_exist"], obj=base_context)

    # THEN the command exits with a non-zero exit code
    assert result.exit_code == EXIT_FAIL


def test_get_sequencing_run_required(
    cli_runner: CliRunner,
    new_demultiplex_context: CGConfig,
    novaseq_x_flow_cell_id: str,
):
    """Test to get a sequencing run using only the required arguments."""
    # GIVEN a database with a sequencing run

    # WHEN getting the sequencing run
    result = cli_runner.invoke(
        get, ["sequencing-run", novaseq_x_flow_cell_id], obj=new_demultiplex_context
    )

    # THEN the command exits successfully
    assert result.exit_code == EXIT_SUCCESS


def test_get_sequencing_run_output(
    cli_runner: CliRunner,
    new_demultiplex_context: CGConfig,
    novaseq_x_flow_cell_id: str,
):
    """Test that the output of the get sequencing-run command has the correct data."""
    # GIVEN a database with a sequencing run
    status_db: Store = new_demultiplex_context.status_db
    sequencing_run: IlluminaSequencingRun = (
        status_db.get_illumina_sequencing_run_by_device_internal_id(novaseq_x_flow_cell_id)
    )

    # WHEN getting the sequencing run
    result = cli_runner.invoke(
        get, ["sequencing-run", novaseq_x_flow_cell_id], obj=new_demultiplex_context
    )

    # THEN the command exits successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN the sequencing run data is printed to the console
    assert sequencing_run.device.internal_id in result.output
    assert sequencing_run.sequencer_type in result.output
    assert sequencing_run.sequencer_name in result.output
    assert str(sequencing_run.sequencing_started_at.date()) in result.output
    assert str(sequencing_run.sequencing_completed_at.date()) in result.output
    assert sequencing_run.data_availability in result.output


def test_get_sequencing_run_archived_at_none(
    cli_runner: CliRunner,
    new_demultiplex_context: CGConfig,
    novaseq_x_flow_cell_id: str,
):
    """Test the output when there is no archived_at date."""
    # GIVEN a non-archived Illumina sequencing run
    status_db: Store = new_demultiplex_context.status_db
    sequencing_run: IlluminaSequencingRun = (
        status_db.get_illumina_sequencing_run_by_device_internal_id(novaseq_x_flow_cell_id)
    )
    sequencing_run.archived_at = None

    # WHEN getting a sequencing run
    result = cli_runner.invoke(
        get, ["sequencing-run", novaseq_x_flow_cell_id], obj=new_demultiplex_context
    )

    # THEN the command exits successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN log that the sequencing run is not archived
    assert "No" in result.output


def test_get_sequencing_run_archived_at_date(
    cli_runner: CliRunner,
    new_demultiplex_context: CGConfig,
    novaseq_x_flow_cell_id: str,
    timestamp: datetime,
):
    """Test that the output has the run's archived at date."""
    # GIVEN a database with an archived sequencing run
    status_db: Store = new_demultiplex_context.status_db
    sequencing_run: IlluminaSequencingRun = (
        status_db.get_illumina_sequencing_run_by_device_internal_id(novaseq_x_flow_cell_id)
    )
    sequencing_run.archived_at = timestamp

    # WHEN getting the sequencing run
    result = cli_runner.invoke(
        get, ["sequencing-run", novaseq_x_flow_cell_id], obj=new_demultiplex_context
    )

    # THEN the command exits successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN the archived time should be in the output
    assert str(timestamp.date()) in result.output


def test_get_sequencing_run_samples_without_samples(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers, caplog
):
    """Test the output when there are no samples associated with the run."""
    # GIVEN a database with a sequencing run without related samples
    status_db: Store = base_context.status_db
    flow_cell: IlluminaFlowCell = helpers.ensure_illumina_flow_cell(store=status_db)
    helpers.ensure_illumina_sequencing_run(store=status_db, flow_cell=flow_cell)

    # WHEN getting a sequencing run with the --samples flag
    result = cli_runner.invoke(
        get,
        ["sequencing-run", flow_cell.internal_id, "--samples"],
        obj=base_context,
    )

    # THEN the command exits successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN a message about no samples should have been displayed
    assert "No samples found on sequencing run" in caplog.text


def test_get_sequencing_run_samples(
    cli_runner: CliRunner, new_demultiplex_context: CGConfig, novaseq_x_flow_cell_id: str
):
    """Test that the output has the samples of the sequencing run."""
    # GIVEN a database with a sequencing run with related samples
    status_db: Store = new_demultiplex_context.status_db
    sequencing_run: IlluminaSequencingRun = (
        status_db.get_illumina_sequencing_run_by_device_internal_id(novaseq_x_flow_cell_id)
    )

    # WHEN getting a sequencing run with the --samples flag
    result = cli_runner.invoke(
        get,
        ["sequencing-run", novaseq_x_flow_cell_id, "--samples"],
        obj=new_demultiplex_context,
    )

    # THEN the command exits successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN all related samples should be listed in the output
    for metric in sequencing_run.sample_metrics:
        assert metric.sample.internal_id in result.output


def test_get_sequencing_run_no_samples(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test that a getting a sequencing run without samples prints no samples."""
    # GIVEN a database with a flow cell without related samples
    status_db: Store = base_context.status_db
    flow_cell: IlluminaFlowCell = helpers.ensure_illumina_flow_cell(store=status_db)
    helpers.ensure_illumina_sequencing_run(store=status_db, flow_cell=flow_cell)

    # WHEN getting a sequencing run with the --no-samples flag
    result = cli_runner.invoke(
        get, ["sequencing-run", flow_cell.internal_id, "--no-samples"], obj=base_context
    )

    # THEN the command exits successfully
    assert result.exit_code == EXIT_SUCCESS


def test_get_flow_cell_no_samples_with_samples(
    cli_runner: CliRunner, new_demultiplex_context: CGConfig, novaseq_x_flow_cell_id: str
):
    """Test that the output has the data of the flow cell."""
    # GIVEN a database with a sequencing run with related samples
    status_db: Store = new_demultiplex_context.status_db
    sequencing_run: IlluminaSequencingRun = (
        status_db.get_illumina_sequencing_run_by_device_internal_id(novaseq_x_flow_cell_id)
    )

    # WHEN getting a sequencing run with the --no-samples flag
    result = cli_runner.invoke(
        get, ["sequencing-run", novaseq_x_flow_cell_id, "--no-samples"], obj=new_demultiplex_context
    )

    # THEN the command exits successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN no related samples should be listed in the output
    for metric in sequencing_run.sample_metrics:
        assert metric.sample.internal_id not in result.output



from cg.cli.workflow.raw_data.base import store_raw_data_analysis
from cg.store.models import Analysis, Case


def test_store_raw_data_analysis(another_case_id: str, cli_runner, raw_data_fastq_context, helpers):
    """Test for CLI command creating an analysis object for a fastq case"""

    # GIVEN a raw data fastq context
    helpers.ensure_case(raw_data_fastq_context.status_db, case_id=another_case_id)
    case_obj: Case = raw_data_fastq_context.status_db.get_case_by_internal_id(another_case_id)
    assert not case_obj.analyses

    # WHEN a command is run to create an analysis for the case
    cli_runner.invoke(store_raw_data_analysis, [another_case_id], obj=raw_data_fastq_context)

    # THEN the analysis is created
    assert (
        len(
            raw_data_fastq_context.status_db._get_query(table=Analysis)
            .filter(Analysis.case_id == case_obj.id)
            .all()
        )
        > 0
    )
